import numpy as np
from tqdm import tqdm
from multiprocessing.pool import Pool

from . import io_tools
from . shared import check_for_boundary_surface_potentials
from .. params import parameters
from .. params.parameters import BinaryInitialParameters
from .. models import lc as lc_model
from .. models import serializers
from ... import units as u
from ... import settings
from ... utils import split_to_batches
from ... logger import getLogger
from ... binary_system.surface.gravity import calculate_polar_gravity_acceleration
from ... base.error import MaxIterationError, MorphologyError
from ... binary_system.curves.community import RadialVelocitySystem


logger = getLogger('analytics.binary_fit.summary')

DASH_N = 126


def fit_lc_summary_with_error_propagation(fit_instance, path, percentiles, dimensionless_radii=True):
    """
    Function propagates errors of fitted parameters during MCMC fit by evaluating the corresponding chain.

    :param fit_instance: elisa.analytics.binary_fit.lc_fit.LCFit
    :param path: str; results will be written here
    :param percentiles: List; [bottom, middle, top] percentiles used for the creation of confidence intervals
    :param dimensionless_radii: if true, radii are provided in SMA units
    :return: None;
    """
    def propagate_errors(radius, factor):
        rel_rad = radius[1:] / radius[0]
        rel_factor = factor[1:] / factor[0]
        retval = np.empty(radius.shape)
        retval[0] = radius[0] * factor[0]
        retval[1:] = retval[0] * (rel_factor + rel_rad)
        return retval

    f = None
    if path is not None:
        f = open(path, 'w')
        write_fn = f.write
        line_sep = '\n'
    else:
        write_fn = print
        line_sep = ''

    flat_chain = fit_instance.flat_chain
    variable_labels = fit_instance.variable_labels
    normalization = fit_instance.normalization

    flat_params = fit_instance.flat_result
    params = fit_instance.result

    params_instance = BinaryInitialParameters(**params)
    setattr(fit_instance.fit_method_instance, 'fixed', params_instance.get_fixed(jsonify=False))
    setattr(fit_instance.fit_method_instance, 'constrained', params_instance.get_constrained(jsonify=False))
    setattr(fit_instance.fit_method_instance, 'fitable', params_instance.get_fitable(jsonify=False))

    if flat_chain is None:
        raise ValueError('MCMC chain was not found.')

    renormalized_chain = np.empty(flat_chain.shape)
    for ii, lbl in enumerate(variable_labels):
        renormalized_chain[:, ii] = parameters.renormalize_value(flat_chain[:, ii], normalization[lbl][0],
                                                                 normalization[lbl][1])
    stop_idx = {}
    # creating param list of the binary system
    complete_param_list = [
        'system@mass_ratio', 'system@semi_major_axis', 'system@asini', 'system@inclination', 'system@eccentricity',
        'system@argument_of_periastron', 'system@gamma', 'system@period', 'system@primary_minimum_time',
        'system@additional_light', 'system@phase_shift']

    component_param_list = [
        'mass', 'surface_potential', 'synchronicity', 'equivalent_radius',
        'polar_radius', 'backward_radius', 'side_radius', 'forward_radius', 't_eff',
        'gravity_darkening', 'albedo', 'metallicity', 'critical_potential', 'polar_log_g', 'bolometric_luminosity'
    ]

    stop_idx['system'] = len(complete_param_list)
    spot_numbers, pulsation_numbers = {}, {}
    spot_lbls, pulse_lbls = {}, {}

    for component in settings.BINARY_COUNTERPARTS.keys():
        complete_param_list += [f'{component}@{prm}' for prm in component_param_list]

        stop_idx[f'{component}_params'] = len(complete_param_list)
        spot_lbls[component] = set([lbl.split('@')[2] for lbl in flat_params.keys() if f'{component}@spot' in lbl])
        spot_numbers[component] = len(spot_lbls[component])
        for spot in spot_lbls[component]:
            complete_param_list += [
                f'{component}@spot@{spot}@longitude', f'{component}@spot@{spot}@latitude',
                f'{component}@spot@{spot}@angular_radius', f'{component}@spot@{spot}@temperature_factor',
            ]
        stop_idx[f'{component}_spots'] = len(complete_param_list)
        pulse_lbls[component] = \
            set([lbl.split('@')[2] for lbl in flat_params.keys() if f'{component}@pulsation' in lbl])
        pulsation_numbers[component] = len(pulse_lbls[component])
        for mode in pulse_lbls[component]:
            complete_param_list += [
                f'{component}@pulsation@{mode}@l', f'{component}@pulsation@{mode}@m',
                f'{component}@pulsation@{mode}@amplitude', f'{component}@pulsation@{mode}@frequency',
                f'{component}@pulsation@{mode}@start_phase', f'{component}@pulsation@{mode}@mode_axis_phi',
                f'{component}@pulsation@{mode}@mode_axis_theta',
            ]
        stop_idx[f'{component}_pulsations'] = len(complete_param_list)

    param_columns = {lbl: ii for ii, lbl in enumerate(complete_param_list)}

    # obtaining binary parameters for each item in MCMC chain
    args = (fit_instance, param_columns, stop_idx, spot_numbers, pulsation_numbers, len(component_param_list))
    full_chain = _manage_chain_evaluation(renormalized_chain, evaluate_binary_params, *args)

    full_chain_mask = (~np.isnan(full_chain)).any(axis=1)
    full_chain = full_chain[full_chain_mask]

    # avoiding np warnings about NANs
    calculated_percentiles = np.empty((3, full_chain.shape[1]))
    calculated_percentiles[:] = np.NaN
    full_chain_mask = (~np.isnan(full_chain)).any(axis=0)

    # evaluating posterior distribution of each binary parameter
    calculated_percentiles[:, full_chain_mask] = np.percentile(full_chain[:, full_chain_mask], percentiles, axis=0)
    full_chain_results = np.row_stack((calculated_percentiles[1, :],
                                       calculated_percentiles[1, :] - calculated_percentiles[0, :],
                                       calculated_percentiles[2, :] - calculated_percentiles[1, :]))

    # output to screen/file
    intro = (write_fn, 'Parameter', 'value', '-1 sigma', '+1 sigma', 'unit', 'status', line_sep)
    write_fn(f"\nBINARY SYSTEM{line_sep}")
    io_tools.write_ln(*intro)
    write_fn(f"{'-' * DASH_N}{line_sep}")

    io_tools.write_propagated_ln(full_chain_results[:, param_columns['system@mass_ratio']], flat_params,
                                 'system@mass_ratio', 'Mass ratio (q=M_2/M_1):', write_fn, line_sep, '-')

    sma_value = full_chain_results[:, param_columns['system@semi_major_axis']]
    sma = (sma_value * u.DefaultBinarySystemUnits.system.semi_major_axis).\
        to(u.DefaultBinarySystemInputUnits.system.semi_major_axis).value
    sma_factor = sma if not dimensionless_radii else np.array([1.0, 0.0, 0.0])
    sma_unit = 'solRad' if not dimensionless_radii else 'SMA'
    io_tools.write_propagated_ln(sma, flat_params, 'system@semi_major_axis', 'Semi major axis (a):', write_fn, line_sep,
                                 'solRad')

    asini_value = full_chain_results[:, param_columns['system@asini']]
    asini = (asini_value * u.DefaultBinarySystemUnits.system.semi_major_axis).\
        to(u.DefaultBinarySystemInputUnits.system.semi_major_axis).value
    io_tools.write_propagated_ln(asini, flat_params, 'system@asini', 'a*sin(i):', write_fn,
                                 line_sep, 'solRad')

    incl = (full_chain_results[:, param_columns['system@inclination']] *
            u.DefaultBinarySystemUnits.system.inclination).to(u.deg).value
    io_tools.write_propagated_ln(incl, flat_params, 'system@inclination', 'Inclination (i):', write_fn, line_sep, 'deg')

    io_tools.write_propagated_ln(full_chain_results[:, param_columns['system@eccentricity']], flat_params,
                                 'system@eccentricity', 'Eccentricity (e):', write_fn, line_sep, '-')

    omega = (full_chain_results[:, param_columns['system@argument_of_periastron']] *
             u.DefaultBinarySystemUnits.system.argument_of_periastron).to(u.deg).value
    io_tools.write_propagated_ln(omega, flat_params, 'system@argument_of_periastron', 'Argument of periastron (omega):',
                                 write_fn, line_sep, 'deg')

    gamma = (full_chain_results[:, param_columns['system@gamma']] *
             u.DefaultBinarySystemUnits.system.gamma).to(u.km / u.s).value
    io_tools.write_propagated_ln(gamma, flat_params, 'system@gamma', 'Centre of mass velocity (gamma):',
                                 write_fn, line_sep, 'km/s')

    io_tools.write_propagated_ln(full_chain_results[:, param_columns['system@period']], flat_params,
                                 'system@period', 'Orbital period (P):', write_fn, line_sep, 'd')

    if 'system@primary_minimum_time' in flat_params:
        io_tools.write_param_ln(flat_params, 'system@primary_minimum_time', 'Time of primary minimum (T0):',
                                write_fn, line_sep)

    io_tools.write_propagated_ln(full_chain_results[:, param_columns['system@additional_light']], flat_params,
                                 'system@additional_light', 'Additional light (l_3):', write_fn,
                                 line_sep, '-')

    io_tools.write_propagated_ln(full_chain_results[:, param_columns['system@phase_shift']], flat_params,
                                 'system@phase_shift', 'Phase shift:', write_fn,
                                 line_sep, '-')

    write_fn(f"{'-' * DASH_N}{line_sep}")

    for component in settings.BINARY_COUNTERPARTS.keys():
        comp_n = 1 if component == 'primary' else 2
        write_fn(f"{component.upper()} COMPONENT{line_sep}")
        io_tools.write_ln(*intro)
        write_fn(f"{'-' * DASH_N}{line_sep}")

        mass = (full_chain_results[:, param_columns[f'{component}@mass']] * u.DefaultBinarySystemUnits.component.mass).\
            to(u.solMass).value
        io_tools.write_propagated_ln(mass, flat_params, f'{component}@mass', f'Mass (M_{comp_n}):', write_fn, line_sep,
                                     'solMas')

        io_tools.write_propagated_ln(full_chain_results[:, param_columns[f'{component}@surface_potential']],
                                     flat_params,
                                     f'{component}@surface_potential', f'Surface potential (Omega_{comp_n}):',
                                     write_fn, line_sep, '-')
        io_tools.write_propagated_ln(full_chain_results[:, param_columns[f'{component}@critical_potential']],
                                     flat_params, f'{component}@critical_potential', 'Critical potential at L_1:',
                                     write_fn, line_sep, '-')
        io_tools.write_propagated_ln(full_chain_results[:, param_columns[f'{component}@synchronicity']],
                                     flat_params, f'{component}@synchronicity', f'Synchronicity (F_{comp_n}):',
                                     write_fn, line_sep, '-')
        io_tools.write_propagated_ln(full_chain_results[:, param_columns[f'{component}@polar_log_g']],
                                     flat_params, f'{component}@polar_log_g', 'Polar gravity (log g):',
                                     write_fn, line_sep, 'log(cgs)')

        r_equiv = propagate_errors(full_chain_results[:, param_columns[f'{component}@equivalent_radius']], sma_factor)
        r_polar = propagate_errors(full_chain_results[:, param_columns[f'{component}@polar_radius']], sma_factor)
        r_backw = propagate_errors(full_chain_results[:, param_columns[f'{component}@backward_radius']], sma_factor)
        r_side = propagate_errors(full_chain_results[:, param_columns[f'{component}@side_radius']], sma_factor)

        io_tools.write_propagated_ln(r_equiv, flat_params, f'{component}@equivalent_radius',
                                     'Equivalent radius (R_equiv):', write_fn, line_sep, sma_unit)
        write_fn(f"\nPeriastron radii{line_sep}")

        io_tools.write_propagated_ln(r_polar, flat_params, f'{component}@polar_radius',
                                     'Polar radius:', write_fn, line_sep, sma_unit)
        io_tools.write_propagated_ln(r_backw, flat_params, f'{component}@backw_radius',
                                     'Backward radius:', write_fn, line_sep, sma_unit)
        io_tools.write_propagated_ln(r_side, flat_params, f'{component}@side_radius',
                                     'Side radius:', write_fn, line_sep, sma_unit)
        if fit_instance.morphology != 'over-contact':
            r_forw = propagate_errors(full_chain_results[:, param_columns[f'{component}@forward_radius']], sma_factor)
            io_tools.write_propagated_ln(r_forw, flat_params, f'{component}@forward_radius',
                                         'Forward radius:', write_fn, line_sep, sma_unit)

        write_fn(f"\nAtmospheric parameters{line_sep}")
        io_tools.write_propagated_ln(full_chain_results[:, param_columns[f'{component}@t_eff']],
                                     flat_params, f'{component}@t_eff', f'Effective temperature (T_eff{comp_n}):',
                                     write_fn, line_sep, 'K')

        l_bol = (full_chain_results[:, param_columns[f'{component}@bolometric_luminosity']] *
                 u.LUMINOSITY_UNIT).to('L_sun').value
        io_tools.write_propagated_ln(l_bol, flat_params, f'{component}@bolometric_luminosity',
                                     'Bolometric luminosity (L_bol): ', write_fn, line_sep, 'L_sol')
        io_tools.write_propagated_ln(full_chain_results[:, param_columns[f'{component}@gravity_darkening']],
                                     flat_params, f'{component}@gravity_darkening',
                                     f'Gravity darkening factor (G_{comp_n}):', write_fn, line_sep, '-')
        io_tools.write_propagated_ln(full_chain_results[:, param_columns[f'{component}@albedo']],
                                     flat_params, f'{component}@albedo',
                                     f'Albedo (A_{comp_n}):', write_fn, line_sep, '-')
        io_tools.write_propagated_ln(full_chain_results[:, param_columns[f'{component}@metallicity']],
                                     flat_params, f'{component}@metallicity',
                                     'Metallicity (log10(X_Fe/X_H)):', write_fn, line_sep, '-')

        if spot_numbers[component] > 0:
            write_fn(f"{'-' * DASH_N}{line_sep}")
            write_fn(f"{component.upper()} SPOTS{line_sep}")
            io_tools.write_ln(*intro)

            for spot_lbl in spot_lbls[component]:
                longitude = (full_chain_results[:, param_columns[f'{component}@spot@{spot_lbl}@longitude']] *
                             u.DefaultSpotUnits.longitude).to(u.deg).value
                io_tools.write_propagated_ln(longitude, flat_params, f'{component}@spot@{spot_lbl}@longitude',
                                             'Longitude: ', write_fn, line_sep, 'deg')
                latitude = (full_chain_results[:, param_columns[f'{component}@spot@{spot_lbl}@latitude']] *
                            u.DefaultSpotUnits.latitude).to(u.deg).value
                io_tools.write_propagated_ln(latitude, flat_params, f'{component}@spot@{spot_lbl}@latitude',
                                             'Latitude: ', write_fn, line_sep, 'deg')
                ang_rad = (full_chain_results[:, param_columns[f'{component}@spot@{spot_lbl}@angular_radius']] *
                           u.DefaultSpotUnits.angular_radius).to(u.deg).value
                io_tools.write_propagated_ln(ang_rad, flat_params, f'{component}@spot@{spot_lbl}@angular_radius',
                                             'Angular radius: ', write_fn, line_sep, 'deg')
                io_tools.write_propagated_ln(
                    full_chain_results[:, param_columns[f'{component}@spot@{spot_lbl}@temperature_factor']],
                    flat_params, f'{component}@spot@{spot_lbl}@temperature_factor',
                    'Temperature factor (T_spot/T_eff): ', write_fn, line_sep, '-'
                )

        if pulsation_numbers[component] > 0:
            write_fn(f"{'-' * DASH_N}{line_sep}")
            write_fn(f"{component.upper()} PULSATIONS{line_sep}")
            io_tools.write_ln(*intro)

            for pulse_lbl in pulse_lbls[component]:
                io_tools.write_propagated_ln(
                    full_chain_results[:, param_columns[f'{component}@pulsation@{pulse_lbl}@l']],
                    flat_params, f'{component}@pulsation@{pulse_lbl}@l', 'Angular degree (l): ', write_fn, line_sep,
                    '-'
                )
                io_tools.write_propagated_ln(
                    full_chain_results[:, param_columns[f'{component}@pulsation@{pulse_lbl}@m']],
                    flat_params, f'{component}@pulsation@{pulse_lbl}@m', 'Azimuthal order (m):  ', write_fn,
                    line_sep, '-'
                )
                io_tools.write_propagated_ln(
                    full_chain_results[:, param_columns[f'{component}@pulsation@{pulse_lbl}@amplitude']],
                    flat_params, f'{component}@pulsation@{pulse_lbl}@m', 'Amplitude (A): ', write_fn, line_sep, '-'
                )
                io_tools.write_propagated_ln(
                    full_chain_results[:, param_columns[f'{component}@pulsation@{pulse_lbl}@frequency']],
                    flat_params, f'{component}@pulsation@{pulse_lbl}@frequency', 'Frequency (f): ', write_fn,
                    line_sep, 'c/d'
                )
                io_tools.write_propagated_ln(
                    full_chain_results[:, param_columns[f'{component}@pulsation@{pulse_lbl}@start_phase']],
                    flat_params, f'{component}@pulsation@{pulse_lbl}@start_phase', 'Initial phase (at T_0): ',
                    write_fn, line_sep, '-'
                )
                mode_phi = (full_chain_results[:,
                            param_columns[f'{component}@spot@{pulse_lbl}@mode_axis_phi']] *
                            u.DefaultPulsationsUnits.mode_axis_phi).to(u.deg).value
                io_tools.write_propagated_ln(
                    mode_phi, flat_params, f'{component}@pulsation@{pulse_lbl}@mode_axis_phi',
                    'Longitude of mode axis: ', write_fn, line_sep, 'deg'
                )
                mode_theta = (full_chain_results[:,
                              param_columns[f'{component}@spot@{pulse_lbl}@mode_axis_theta']] *
                              u.DefaultPulsationsUnits.mode_axis_theta).to(u.deg).value
                io_tools.write_propagated_ln(
                    mode_theta, flat_params, f'{component}@pulsation@{pulse_lbl}@mode_axis_theta',
                    'Latitude of mode axis: ', write_fn, line_sep, 'deg'
                )

        write_fn(f"{'-' * DASH_N}{line_sep}")

    if f is not None:
        f.close()


def evaluate_binary_params(*args):
    """
    Evaluates MCMC (sub)chain for error propagation in fit_summary function.

    :param args: Tuple;

    ::

        Tuple[
                fit_instance: elisa.analytics.binary_fit.lc_fit.LCFit;
                param_columns: Dict; name of the column: index of the column,
                stop_idx: Dict; indices of endpoints for certain parameter groups eg. system, primary parameters,
                                primary spots etc...,
                spot_numbers: Dict; number of spots for each components,
                pulsation_numbers: Dict; number of pulsation modes for each components,
                cpl: int; component parameter
                renormalized_chain: numpy.array; mcmc chain with renormalized (actual) values
            ]
    """
    fit_instance, param_columns, stop_idx, spot_numbers, pulsation_numbers, cpl, renormalized_chain = args
    full_chain = np.empty((renormalized_chain.shape[0], len(param_columns)))
    full_chain[:] = np.nan

    for ii in tqdm(range(renormalized_chain.shape[0])):
        init_binary_kwargs = parameters.prepare_properties_set(renormalized_chain[ii, :],
                                                               fit_instance.fit_method_instance.fitable.keys(),
                                                               fit_instance.fit_method_instance.constrained,
                                                               fit_instance.fit_method_instance.fixed)

        try:
            binary_instance = lc_model.prepare_binary(_verify=False, **init_binary_kwargs)
            for var_label in list(param_columns.keys())[:stop_idx['system']]:
                full_chain[ii, param_columns[var_label]] = getattr(binary_instance, var_label.split('@')[1], None)

            for component in settings.BINARY_COUNTERPARTS.keys():
                star_instance = getattr(binary_instance, component)
                for var_label in list(param_columns.keys())[stop_idx[f'{component}_params'] -
                                                            cpl:stop_idx[f'{component}_params'] - 3]:
                    full_chain[ii, param_columns[var_label]] = getattr(star_instance, var_label.split('@')[1], None)

                polar_g = calculate_polar_gravity_acceleration(star_instance,
                                                               1 - binary_instance.eccentricity,
                                                               binary_instance.mass_ratio,
                                                               component=component,
                                                               semi_major_axis=binary_instance.semi_major_axis,
                                                               synchronicity=star_instance.synchronicity,
                                                               logg=True) + 2
                full_chain[ii, param_columns[f'{component}@polar_log_g']] = polar_g

                crit_pot = binary_instance.critical_potential(component, 1.0 - binary_instance.eccentricity)
                full_chain[ii, param_columns[f'{component}@critical_potential']] = crit_pot

                l_bol = binary_instance.calculate_bolometric_luminosity(components=component)[component]
                full_chain[ii, param_columns[f'{component}@bolometric_luminosity']] = l_bol

                if spot_numbers[component] == 0 and pulsation_numbers[component] == 0:
                    continue

                ref_idx = stop_idx[f'{component}_params']
                for spot_idx in range(spot_numbers[component]):
                    full_chain[ii, ref_idx + 4 * spot_idx] = star_instance.spots[spot_idx].longitude
                    full_chain[ii, ref_idx + 4 * spot_idx + 1] = star_instance.spots[spot_idx].latitude
                    full_chain[ii, ref_idx + 4 * spot_idx + 2] = star_instance.spots[spot_idx].angular_radius
                    full_chain[ii, ref_idx + 4 * spot_idx + 3] = star_instance.spots[spot_idx].temperature_factor

                ref_idx = stop_idx[f'{component}_spots']
                for pulse_idx in range(pulsation_numbers[component]):
                    full_chain[ii, ref_idx + 7*pulse_idx] = star_instance.pulsations[pulse_idx].l
                    full_chain[ii, ref_idx + 7*pulse_idx + 1] = star_instance.pulsations[pulse_idx].m
                    full_chain[ii, ref_idx + 7*pulse_idx + 2] = star_instance.pulsations[pulse_idx].amplitude
                    full_chain[ii, ref_idx + 7*pulse_idx + 3] = star_instance.pulsations[pulse_idx].frequency
                    full_chain[ii, ref_idx + 7*pulse_idx + 4] = star_instance.pulsations[pulse_idx].start_phase
                    full_chain[ii, ref_idx + 7*pulse_idx + 5] = star_instance.pulsations[pulse_idx].mode_axis_phi
                    full_chain[ii, ref_idx + 7*pulse_idx + 6] = star_instance.pulsations[pulse_idx].mode_axis_theta

        except (MaxIterationError, MorphologyError):
            continue

    return full_chain


def simple_lc_fit_summary(fit_instance, path, dimensionless_radii=True):
    """
    Function returns or saves to file a comprehensive report on given fit procedure. Summary contains all fitted or
    derived binary parameters.

    :param fit_instance: LCFit instance
    :param path: str; file in which to store a summary
    :param dimensionless_radii: if true, radii are provided in SMA units
    """
    f = None
    if path is not None:
        f = open(path, 'w')
        write_fn = f.write
        line_sep = '\n'
    else:
        write_fn = print
        line_sep = ''

    try:
        result_dict = fit_instance.flat_result
        result_dict = check_for_boundary_surface_potentials(result_dict, fit_instance.morphology)
        b_kwargs = {key: val['value'] for key, val in result_dict.items()}
        binary_instance = lc_model.prepare_binary(_verify=False, **b_kwargs)

        intro = (write_fn, 'Parameter', 'value', '-1 sigma', '+1 sigma', 'unit', 'status', line_sep)
        write_fn(f"\nBINARY SYSTEM{line_sep}")
        io_tools.write_ln(*intro)
        write_fn(f"{'-' * DASH_N}{line_sep}")

        q_desig = 'Mass ratio (q=M_2/M_1):'
        a_desig = 'Semi major axis (a):'

        if 'system@mass_ratio' in result_dict:
            io_tools.write_param_ln(result_dict, 'system@mass_ratio', q_desig, write_fn, line_sep, 3)
            io_tools.write_param_ln(result_dict, 'system@semi_major_axis', a_desig, write_fn, line_sep, 3)
            sma_factor = result_dict['system@semi_major_axis']['value'] if not dimensionless_radii else 1.0
            sma_unit = str(result_dict['system@semi_major_axis']['unit']) if not dimensionless_radii else 'SMA'
        else:
            io_tools.write_ln(write_fn, q_desig, binary_instance.mass_ratio, '', '', '', 'derived', line_sep, 3)
            sma = (binary_instance.semi_major_axis * u.DefaultBinarySystemUnits.system.semi_major_axis).\
                to(u.DefaultBinarySystemInputUnits.system.semi_major_axis).value
            io_tools.write_ln(write_fn, a_desig, sma, '', '', 'solRad', 'derived', line_sep, 3)
            sma_factor = sma
            sma_unit = 'solRad' if not dimensionless_radii else 'SMA'

        io_tools.write_param_ln(result_dict, 'system@inclination', 'Inclination (i):', write_fn, line_sep, 2)
        io_tools.write_param_ln(result_dict, 'system@eccentricity', 'Eccentricity (e):', write_fn, line_sep, 2)
        io_tools.write_param_ln(result_dict, 'system@argument_of_periastron', 'Argument of periastron (omega):',
                                write_fn, line_sep, 2)
        io_tools.write_param_ln(result_dict, 'system@period', 'Orbital period (P):', write_fn, line_sep)

        if 'system@primary_minimum_time' in result_dict:
            io_tools.write_param_ln(result_dict, 'system@primary_minimum_time', 'Time of primary minimum (T0):',
                                    write_fn, line_sep)
        if 'system@additional_light' in result_dict:
            io_tools.write_param_ln(result_dict, 'system@additional_light', 'Additional light (l_3):',
                                    write_fn, line_sep, 4)

        p_desig = 'Phase shift:'
        if 'system@phase_shift' in result_dict:
            io_tools.write_param_ln(result_dict, 'system@phase_shift', p_desig, write_fn, line_sep)
        else:
            io_tools.write_ln(write_fn, p_desig, 0.0, '-', '-', '-', 'Fixed', line_sep)

        write_fn(f"{'-' * DASH_N}{line_sep}")

        for component in settings.BINARY_COUNTERPARTS:
            comp_n = 1 if component == 'primary' else 2
            star_instance = getattr(binary_instance, component)
            write_fn(f"{component.upper()} COMPONENT{line_sep}")
            io_tools.write_ln(*intro)
            write_fn(f"{'-' * DASH_N}{line_sep}")

            m_desig = f'Mass (M_{comp_n}):'
            if f'{component}@mass' in result_dict:
                io_tools.write_param_ln(result_dict, f'{component}@mass', m_desig, write_fn, line_sep, 3)
            else:
                mass = (star_instance.mass * u.DefaultBinarySystemUnits.component.mass).to(u.solMass).value
                io_tools.write_ln(write_fn, m_desig, mass,
                                  '-', '-', 'solMass', 'Derived', line_sep, 3)

            io_tools.write_param_ln(result_dict, f'{component}@surface_potential',
                                    f'Surface potential (Omega_{comp_n}):', write_fn, line_sep, 4)

            crit_pot = binary_instance.critical_potential(component, 1.0 - binary_instance.eccentricity)
            io_tools.write_ln(write_fn, 'Critical potential at L_1:', crit_pot, '-', '-', '-', 'Derived', line_sep, 4)

            f_desig = f'Synchronicity (F_{comp_n}):'

            if f'{component}@synchronicity' in result_dict:
                io_tools.write_param_ln(result_dict, f'{component}@synchronicity', f_desig, write_fn, line_sep, 3)
            else:
                io_tools.write_ln(write_fn, f_desig, star_instance.synchronicity, '-', '-', '-', 'Fixed', line_sep, 3)

            polar_g = calculate_polar_gravity_acceleration(star_instance,
                                                           1 - binary_instance.eccentricity,
                                                           binary_instance.mass_ratio,
                                                           component=component,
                                                           semi_major_axis=binary_instance.semi_major_axis,
                                                           synchronicity=star_instance.synchronicity,
                                                           logg=True) + 2

            io_tools.write_ln(write_fn, 'Polar gravity (log g):', polar_g, '-', '-', 'log(cgs)', 'Derived',
                              line_sep, 3)

            io_tools.write_ln(write_fn, 'Equivalent radius (R_equiv):', star_instance.equivalent_radius*sma_factor,
                              '-', '-', sma_unit, 'Derived', line_sep, 5)

            write_fn(f"\nPeriastron radii{line_sep}")

            io_tools.write_ln(write_fn, 'Polar radius:', star_instance.polar_radius*sma_factor, '-', '-', sma_unit,
                              'Derived', line_sep, 5)
            io_tools.write_ln(write_fn, 'Backward radius:', star_instance.backward_radius*sma_factor, '-', '-',
                              sma_unit, 'Derived', line_sep, 5)
            io_tools.write_ln(write_fn, 'Side radius:', star_instance.side_radius*sma_factor, '-', '-', sma_unit,
                              'Derived', line_sep, 5)

            if getattr(star_instance, 'forward_radius', None) is not None:
                io_tools.write_ln(write_fn, 'Forward radius:', star_instance.forward_radius*sma_factor, '-',
                                  '-', sma_unit, 'Derived', line_sep, 5)

            write_fn(f"\nAtmospheric parameters{line_sep}")

            io_tools.write_param_ln(result_dict, f'{component}@t_eff', f'Effective temperature (T_eff{comp_n}):',
                                    write_fn, line_sep, 0)

            l_bol = (binary_instance.calculate_bolometric_luminosity(components=component)[component] *
                     u.LUMINOSITY_UNIT).to('L_sun').value

            io_tools.write_ln(write_fn, 'Bolometric luminosity (L_bol): ', l_bol,
                              '-', '-', 'L_Sol', 'Derived', line_sep, 2)

            f_desig = f'Gravity darkening factor (G_{comp_n}):'
            if f'{component}@gravity_darkening' in result_dict:
                io_tools.write_param_ln(result_dict, f'{component}@gravity_darkening',
                                        f_desig, write_fn, line_sep, 3)
            else:
                args = (write_fn, f_desig, star_instance.gravity_darkening, '-', '-', '-', 'Derived', line_sep, 3)
                io_tools.write_ln(*args)

            f_desig = f'Albedo (A_{comp_n}):'
            if f'{component}@albedo' in result_dict:
                io_tools.write_param_ln(result_dict, f'{component}@albedo', f_desig,
                                        write_fn, line_sep, 3)
            else:
                args = (write_fn, f_desig, star_instance.albedo, '-', '-', '-', 'Derived', line_sep, 3)
                io_tools.write_ln(*args)

            met_desig = 'Metallicity (log10(X_Fe/X_H)):'
            if f'{component}@metallicity' in result_dict:
                io_tools.write_param_ln(result_dict, f'{component}@metallicity', met_desig, write_fn, line_sep, 2)
            else:
                io_tools.write_ln(write_fn, met_desig, star_instance.metallicity,
                                  '-', '-', '-', 'Fixed', line_sep, 2)

            if star_instance.has_spots():
                write_fn(f"{'-' * DASH_N}{line_sep}")
                write_fn(f"{component.upper()} SPOTS{line_sep}")
                io_tools.write_ln(*intro)

                for spot in fit_instance.result[component]["spots"]:
                    write_fn(f'{"-" * DASH_N}{line_sep}')
                    write_fn(f'Spot: {spot["label"]} {line_sep}')
                    write_fn(f'{"-" * DASH_N}{line_sep}')

                    io_tools.write_param_ln(spot, 'longitude', 'Longitude: ', write_fn, line_sep, 3)
                    io_tools.write_param_ln(spot, 'latitude', 'Latitude: ', write_fn, line_sep, 3)
                    io_tools.write_param_ln(spot, 'angular_radius', 'Angular radius: ', write_fn, line_sep, 3)
                    io_tools.write_param_ln(spot, 'temperature_factor', 'Temperature factor (T_spot/T_eff): ',
                                            write_fn, line_sep, 3)

            if star_instance.has_pulsations():
                write_fn(f"{'-' * DASH_N}{line_sep}")
                write_fn(f"{component.upper()} PULSATIONS{line_sep}")
                io_tools.write_ln(*intro)

                for mode in fit_instance.result[component]["pulsations"]:
                    write_fn(f'{"-" * DASH_N}{line_sep}')
                    write_fn(f'Pulsation: {mode["label"]} {line_sep}')
                    write_fn(f'{"-" * DASH_N}{line_sep}')

                    io_tools.write_param_ln(mode, 'l', 'Angular degree (l): ', write_fn, line_sep, 0)
                    io_tools.write_param_ln(mode, 'm', 'Azimuthal order (m): ', write_fn, line_sep, 0)
                    io_tools.write_param_ln(mode, 'amplitude', 'Amplitude (A): ', write_fn, line_sep, 0)
                    io_tools.write_param_ln(mode, 'frequency', 'Frequency (f): ', write_fn, line_sep, 0)
                    io_tools.write_param_ln(mode, 'start_phase', 'Initial phase (at T_0): ', write_fn, line_sep, 3)
                    io_tools.write_param_ln(mode, 'mode_axis_phi', 'Longitude of mode axis: ',
                                            write_fn, line_sep, 1)
                    io_tools.write_param_ln(mode, 'mode_axis_theta', 'Latitude of mode axis: ',
                                            write_fn, line_sep, 1)

            write_fn(f"{'-' * DASH_N}{line_sep}")

        if result_dict.get('r_squared', False):
            if result_dict['r_squared']['value'] is not None:
                io_tools.write_param_ln(result_dict, 'r_squared', 'Fit R^2: ', write_fn, line_sep, 6)

            write_fn(f"{'-' * DASH_N}{line_sep}")
    finally:
        if f is not None:
            f.close()


def simple_rv_fit_summary(fit_instance, path):
    """
    Function returns or saves to file a comprehensive report on given fit procedure. Summary contains all fitted or
    derived binary parameters.

    :param fit_instance: LCFit instance
    :param path: str; file in which to store a summary
    """
    f = None
    if path is not None:
        f = open(path, 'w')
        write_fn = f.write
        line_sep = '\n'
    else:
        write_fn = print
        line_sep = ''

    try:
        write_fn(f"\n{'-' * DASH_N}{line_sep}")
        io_tools.write_ln(write_fn, 'Parameter', 'value', '-1 sigma', '+1 sigma', 'unit', 'status', line_sep)
        write_fn(f"{'-' * DASH_N}{line_sep}")
        result_dict = fit_instance.flat_result

        if 'system@mass_ratio' in result_dict:
            io_tools.write_param_ln(result_dict, 'system@mass_ratio', 'Mass ratio (q=M_2/M_1):',
                                    write_fn, line_sep, 3)
            io_tools.write_param_ln(result_dict, 'system@asini', 'a*sin(i):', write_fn, line_sep, 2)
        else:
            io_tools.write_param_ln(result_dict, 'primary@mass', 'Primary mass:', write_fn, line_sep, 3)
            io_tools.write_param_ln(result_dict, 'secondary@mass', 'Secondary mass:', write_fn, line_sep, 3)
            io_tools.write_param_ln(result_dict, 'system@inclination', 'Inclination(i):', write_fn, line_sep, 3)

        io_tools.write_param_ln(result_dict, 'system@eccentricity', 'Eccentricity (e):', write_fn, line_sep)
        io_tools.write_param_ln(result_dict, 'system@argument_of_periastron',
                                'Argument of periastron (omega):', write_fn, line_sep)
        io_tools.write_param_ln(result_dict, 'system@gamma', 'Centre of mass velocity (gamma):', write_fn, line_sep)
        io_tools.write_param_ln(result_dict, 'system@period', 'Orbital period (P):', write_fn, line_sep)
        if 'system@primary_minimum_time' in result_dict.keys():
            io_tools.write_param_ln(result_dict, 'system@primary_minimum_time',
                                    'Time of primary minimum (T0):', write_fn, line_sep)

        write_fn(f"{'-' * DASH_N}{line_sep}")

        if result_dict.get('r_squared', False):
            if result_dict['r_squared']['value'] is not None:
                io_tools.write_param_ln(result_dict, 'r_squared', 'Fit R^2: ', write_fn, line_sep, 6)

            write_fn(f"{'-' * DASH_N}{line_sep}")
    finally:
        if f is not None:
            f.close()


def fit_rv_summary_with_error_propagation(fit_instance, path, percentiles):
    """
    Performs error propagation using provided MCMC results in form of flat chain stored in `fit_instance`.

    :param fit_instance: elisa.analytics.binary_fit.rv_fit.RVFit;
    :param path: str; results will be written here
    :param percentiles: List; [bottom, middle, top] percentiles used for the creation of confidence intervals
    """
    if path is not None:
        f = open(path, 'w')
        write_fn = f.write
        line_sep = '\n'
    else:
        write_fn = print
        line_sep = ''

    flat_chain = fit_instance.flat_chain
    variable_labels = fit_instance.variable_labels
    normalization = fit_instance.normalization

    flat_params = fit_instance.flat_result
    params = fit_instance.result

    params_instance = BinaryInitialParameters(**params)
    setattr(fit_instance.fit_method_instance, 'fixed', params_instance.get_fixed(jsonify=False))
    setattr(fit_instance.fit_method_instance, 'constrained', params_instance.get_constrained(jsonify=False))
    setattr(fit_instance.fit_method_instance, 'fitable', params_instance.get_fitable(jsonify=False))

    if flat_chain is None:
        raise ValueError('MCMC chain was not found.')

    renormalized_chain = np.empty(flat_chain.shape)
    for ii, lbl in enumerate(variable_labels):
        renormalized_chain[:, ii] = parameters.renormalize_value(flat_chain[:, ii], normalization[lbl][0],
                                                                 normalization[lbl][1])
    stop_idx = {}

    complete_param_list = [
        'system@mass_ratio', 'system@asini', 'system@eccentricity',
        'system@argument_of_periastron', 'system@gamma', 'system@period', 'system@primary_minimum_time']

    param_columns = {lbl: ii for ii, lbl in enumerate(complete_param_list)}

    args = (fit_instance, param_columns, stop_idx)
    full_chain = _manage_chain_evaluation(renormalized_chain, evaluate_rv_params, *args)

    full_chain_mask = (~np.isnan(full_chain)).any(axis=1)
    full_chain = full_chain[full_chain_mask]

    # avoiding np warnings about NANs
    calculated_percentiles = np.empty((3, full_chain.shape[1]))
    calculated_percentiles[:] = np.NaN
    full_chain_mask = (~np.isnan(full_chain)).any(axis=0)
    calculated_percentiles[:, full_chain_mask] = np.percentile(full_chain[:, full_chain_mask], percentiles, axis=0)
    full_chain_results = np.row_stack((calculated_percentiles[1, :],
                                       calculated_percentiles[1, :] - calculated_percentiles[0, :],
                                       calculated_percentiles[2, :] - calculated_percentiles[1, :]))

    intro = (write_fn, 'Parameter', 'value', '-1 sigma', '+1 sigma', 'unit', 'status', line_sep)
    write_fn(f"\nBINARY SYSTEM{line_sep}")
    io_tools.write_ln(*intro)
    write_fn(f"{'-' * DASH_N}{line_sep}")

    io_tools.write_propagated_ln(full_chain_results[:, param_columns['system@mass_ratio']], flat_params,
                                 'system@mass_ratio', 'Mass ratio (q=M_2/M_1):', write_fn, line_sep, '-')

    # sma = (full_chain_results[:, param_columns['system@asini']] *
    #        u.DISTANCE_UNIT).to(u.solRad).value
    sma = full_chain_results[:, param_columns['system@asini']]
    io_tools.write_propagated_ln(sma, flat_params, 'system@asini', 'a*sin(i):', write_fn,
                                 line_sep, 'solRad')

    io_tools.write_propagated_ln(full_chain_results[:, param_columns['system@eccentricity']], flat_params,
                                 'system@eccentricity', 'Eccentricity (e):', write_fn, line_sep, '-')

    omega = (full_chain_results[:, param_columns['system@argument_of_periastron']] *
             u.DefaultBinarySystemUnits.system.argument_of_periastron).to(u.deg).value
    io_tools.write_propagated_ln(omega, flat_params, 'system@argument_of_periastron', 'Argument of periastron (omega):',
                                 write_fn, line_sep, 'deg')

    omega = (full_chain_results[:, param_columns['system@gamma']] *
             u.DefaultBinarySystemUnits.system.gamma).to(u.km/u.s).value
    io_tools.write_propagated_ln(omega, flat_params, 'system@gamma', 'Centre of mass velocity (gamma):',
                                 write_fn, line_sep, 'km/s')

    io_tools.write_propagated_ln(full_chain_results[:, param_columns['system@period']], flat_params,
                                 'system@period', 'Orbital period (P):', write_fn, line_sep, 'd')

    if 'system@primary_minimum_time' in flat_params:
        io_tools.write_param_ln(flat_params, 'system@primary_minimum_time', 'Time of primary minimum (T0):',
                                write_fn, line_sep)

    write_fn(f"{'-' * DASH_N}{line_sep}")


def evaluate_rv_params(*args):
    """
    Produces full set of model parameters from the MCMC chain which will be later used in the error propagation
    technique.

    :param args: Tuple;

    ::

        Tuple[
                fit_instance: elisa.analytics.binary_fit.rv_fit.RVFit;
                param_columns: Dict; name of the column: index of the column, a full set of model parameters
                stop_idx: Dict; indices of endpoints for certain parameter groups eg. system, primary parameters,
                                primary spots etc...,
                renormalized_chain: numpy.array; mcmc chain with renormalized (actual) values
            ]

    :return: numpy.array; distribution of a full set of model parameters
    """
    fit_instance, param_columns, stop_idx, renormalized_chain = args
    full_chain = np.empty((renormalized_chain.shape[0], len(param_columns)))
    for ii in tqdm(range(renormalized_chain.shape[0])):
        init_rv_kwargs = parameters.prepare_properties_set(renormalized_chain[ii, :],
                                                           fit_instance.fit_method_instance.fitable.keys(),
                                                           fit_instance.fit_method_instance.constrained,
                                                           fit_instance.fit_method_instance.fixed)
        system_kwgs = serializers.serialize_system_kwargs(**init_rv_kwargs)

        binary_instance = RadialVelocitySystem(**RadialVelocitySystem.prepare_json(system_kwgs))
        for var_label in list(param_columns.keys())[:-1]:
            full_chain[ii, param_columns[var_label]] = getattr(binary_instance, var_label.split('@')[1], None)
        full_chain[ii, param_columns['system@primary_minimum_time']] = init_rv_kwargs['system@primary_minimum_time']

    return full_chain


def _manage_chain_evaluation(renormalized_chain, eval_function, *args):
    """
    Function manages evaluations of system parameters from flat chains provided by MCMC.
    Enables the use of multiprocessing.

    :param renormalized_chain: numpy.array; glat chain provided by MCMC
    :param eval_function: callable; functions called to evaluate the chain batches
    :return: numpy.array; array of model parameters
    """
    if settings.NUMBER_OF_MCMC_PROCESSES > 1:
        logger.info("starting multiprocessor workers for error propagation technique")
        batches = split_to_batches(array=renormalized_chain, n_proc=settings.NUMBER_OF_MCMC_PROCESSES)
        pool = Pool(processes=settings.NUMBER_OF_MCMC_PROCESSES)
        result = [pool.apply_async(eval_function, args + (batch,)) for batch in batches]
        pool.close()
        pool.join()
        result = [r.get() for r in result]
        return np.row_stack(result)
    else:
        args += (renormalized_chain,)
        return eval_function(*args)



import json
from abc import ABCMeta
from typing import Union, Dict

from . import transform
from . params import bonds, parameters
from . binary_fit import rv_fit, lc_fit
from . binary_fit.plot import (
    RVPlotMCMC, RVPlotLsqr,
    LCPlotLsqr, LCPlotMCMC
)
from .. import utils
from .. logger import getLogger
from elisa import settings

logger = getLogger('analytics.tasks')


class AnalyticsTask(metaclass=ABCMeta):
    """
    Abstract class defining fitting task.  This structure aims to provide a framework for solving inverse problem for
    one object that embeds observed data and fitting methods and provides unified output from fitting methods along with
    capability to visualize the resulting fit.

    :param name: str; arbitrary name of instance
    :param data: Dict; data to bwe analyzed with the Analytics task instance
    """
    ID = 1
    LS_NAMES = ('least_squares', 'least_squares', 'ls', 'LS')
    MCMC_NAMES = ('mcmc', 'MCMC')
    ALLOWED_METHODS = LS_NAMES + MCMC_NAMES
    MANDATORY_KWARGS = ["data", ]
    OPTIONAL_KWARGS = ["atmosphere_model", "limb_darkening_coefficients"]
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS
    CONSTRAINT_OPERATORS = bonds.ALLOWED_CONSTRAINT_METHODS + bonds.ALLOWED_CONSTRAINT_CHARS
    FIT_CLS = None
    PLOT_CLS = None
    TRANSFORM_PROPERTIES_CLS = None

    def __init__(self, method, name=None, **kwargs):
        self.data = dict()
        self.method = method

        if utils.is_empty(name):
            self.name = str(AnalyticsTask.ID)
            logger.debug(f"name of class instance {self.__class__.__name__} set to {self.name}")
            self.__class__.ID += 1
        else:
            self.name = str(name)

        utils.invalid_kwarg_checker(kwargs, self.ALL_KWARGS, self.__class__)
        utils.check_missing_kwargs(self.MANDATORY_KWARGS, kwargs, instance_of=AnalyticsTask)
        kwargs = self.transform_input(**kwargs)
        self.init_properties(**kwargs)

        logger.debug(f'initializing fitting module in class instance '
                     f'{self.__class__.__name__} / {self.name}.')
        self.fit_cls = self.__class__.FIT_CLS()
        self.plot = self.__class__.PLOT_CLS(instance=self.fit_cls, data=self.data)

    @classmethod
    def validate_method(cls, method):
        """
        Checking if the user supplied the correct name for the optimization method.

        :param method: str; name of the optimization method provided by the user
        :return: Union[None, ValueError];
        """
        if method not in cls.ALLOWED_METHODS:
            raise ValueError(f'Invalid fitting method. Use one of: {", ".join(cls.ALLOWED_METHODS)}')

    def load_result(self, filename, autofill_sma=False):
        """
        Function loads a JSON file containing model parameters and stores it as an attribute of AnalyticsTask fitting
        instance. This is useful if you want to examine already calculated results using functionality provided by the
        AnalyticsTask instances (e.g: LCBinaryAnalyticsTask, RVBinaryAnalyticsTask, etc.). I also returns model
        parameters in standard dict (JSON) format.

        :param filename: str;
        :param autofill_sma; bool; if True, the semi-major axis will be autofilled to fitting
                                   parameters if absent
        :return: Dict; model parameters in a standardized format
        """
        self.fit_cls.load_result(filename, autofill_sma=autofill_sma)
        return self.fit_cls.get_result()

    def save_result(self, filename):
        """
        Save result as JSON file.

        :param filename: str; path to file
        """
        self.fit_cls.save_result(filename)

    def set_result(self, result, autofill_sma=False):
        """
        Set model parameters in dictionary (JSON format) as an attribute of AnalyticsTask fitting instance. This is
        useful if you want to examine already calculated results using functionality provided by the AnalyticsTask
        instances (e.g: LCBinaryAnalyticsTask, RVBinaryAnalyticsTask, etc.).

        :param result: Dict; model parameters in JSON format
        :param autofill_sma; bool; if True, the semi-major axis will be autofilled to fitting
                                   parameters if absent
        """
        self.fit_cls.set_result(result, autofill_sma=autofill_sma)

    def get_result(self):
        """
        Returns model parameters in standard dict (JSON) format.

        :return: Dict; model parameters in a standardized format
        """
        return self.fit_cls.get_result()

    def result_summary(self, filename=None, **kwargs):
        """
        Function produces detailed summary of the current fitting task with the possibility to propagate
        uncertainties of the fitted binary model parameters if MCMC method was used and `propagate_errors` is True.

        :param filename: path where to store summary
        :param kwargs: Dict;
        :**kwargs options for MCMC method**:
            * :propagate_errors: bool; errors of fitted parameters will be propagated to the rest of EB
                                       parameters (takes a while to calculate)
            * :percentiles: List; percentiles used to evaluate confidence intervals from posterior
                                  distribution of EB parameters in MCMC chain . Used only when if
                                  `propagate_errors` is True.
            * :dimensionless_radii: bool; if True (default), radii are provided in SMA, otherwise solRad are used,
                                          available only for light curve fitting

        :return:
        """
        self.fit_cls.fit_summary(filename, **kwargs)

    fit_summary = result_summary

    def load_chain(self, filename, discard=0, percentiles=None):
        """
        Function loads MCMC chain along with auxiliary data from json file created after each MCMC run.

        :param percentiles: List;
        :param self: Union[] instance of fitting cls based on method (mcmc, lsqr) and type(lc, rv)
        :param discard: int; Discard the first discard steps in the chain as burn-in. (default: 0)
        :param filename: str; full name of the json file
        :return: Tuple[numpy.ndarray, List, Dict]; flattened mcmc chain, labels of variables in `flat_chain` columns,
                                                  {var_name: (min_boundary, max_boundary), ...} dictionary of
                                                  boundaries defined by user for each variable needed
                                                  to reconstruct real values from normalized `flat_chain` array
        """

        if self.method not in self.MCMC_NAMES:
            raise ValueError('Load chain method can be used only with mcmc task.')
        self.fit_cls.load_chain(filename, discard, percentiles)
        return self

    def filter_chain(self, **boundaries):
        """
        Filtering MCMC chain down to given parameter intervals. This function is useful in case of bimodal distribution
        of the MCMC chain.

        :param boundaries: Dict; dictionary of boundaries e.g. {'primary@te_ff': (5000, 6000), other parameters ...}
        :return: numpy.array; filtered flat chain
        """
        if self.method not in self.MCMC_NAMES:
            raise ValueError('Filter chain method can be used only with mcmc task.')
        self.fit_cls.filter_chain(**boundaries)

    def fit(self, x0: Union[Dict, parameters.BinaryInitialParameters], **kwargs):
        """
        Function solves an inverse task of inferring parameters of the eclipsing binary from the observed light curve.
        Least squares method is adopted from scipy.optimize.least_squares.
        MCMC uses emcee package to perform sampling.

        :param x0: Dict; initial state of the sampler, model parameters in standard JSON format
        :param kwargs: Dict; method-dependent
        :**additional light curve kwargs**:
            * **morphology** * - str - `detached` or `over-contact`
            * **interp_treshold** * - int - Above this total number of datapoints, light curve will be interpolated
                                            using model containing `interp_treshold` equidistant points per epoch
            * **discretization** * - Union[int, float] - discretization factor of the primary component, default: 5
            * **samples** * - Union[str, List]; 'uniform' (equidistant sampling in phase), 'adaptive'
                                                (equidistant sampling on curve) or list with phases in (0, 1) interval

        :**kwargs options for least_squares**: passes arguments of scipy.optimize.least_squares method except
                                               `fun`, `x0` and `bounds`

        :**kwargs options for MCMC method**:
            * **nwalkers (int)** * - The number of walkers in the ensemble. Minimum is 2 * number of free parameters.
            * **nsteps (int)** * - The number of steps to run. Default is 1000.
            * **initial_state (numpy.ndarray)** * - The initial state or position vector made of free parameters with
                    shape (nwalkers, number of free parameters). The order of the parameters is specified by their order
                    in `x0`. Be aware that initial states should be supplied in normalized form (0, 1). For example, 0
                    means value of the parameter at `min` and 1.0 at `max` value in `x0`. By default, they are generated
                    randomly uniform distribution.
            * **burn_in** * - The expected number of steps to run to achieve equilibrium where useful sampling can
                              start.Default value is nsteps / 10.
            * **progress** * - bool - display the progress bar of the sampling
            * **percentiles** * - List of percentiles used to create error results and confidence interval from MCMC
                                  chain. Default value is [16, 50, 84] (1-sigma confidence interval)
            * **save** * - bool - save chain
            * **fit_id** * - str - identificator or location of stored chain
            * **samples** * - Union[str, List]; 'uniform' (equidistant sampling in phase), 'adaptive'
                                                (equidistant sampling on curve) or list with phases in (0, 1) interval

        :return: Dict; resulting parameters {param_name: {`value`: value, `unit`: astropy.unit, ...}, ...}
        """
        if isinstance(x0, dict):
            x0 = parameters.BinaryInitialParameters(**x0)
        return self.fit_cls.fit(x0, data=self.data, **kwargs)

    def coefficient_of_determination(self, model_parameters=None, discretization=5, interpolation_treshold=None):
        """
        Function returns R^2 for given model parameters and observed data.

        :param model_parameters: Dict; if None, get_result() is called
        :param discretization: float;
        :param interpolation_treshold: int; if None settings.MAX_CURVE_DATA_POINTS is used
        :return: float;
        """
        model_parameters = self.get_result() if model_parameters is None else model_parameters
        interpolation_treshold = settings.MAX_CURVE_DATA_POINTS \
            if interpolation_treshold is None else interpolation_treshold

        r2 = self.fit_cls.coefficient_of_determination(
            model_parameters,
            self.data,
            discretization,
            interpolation_treshold
        )
        model_parameters['r_squared'] = r2
        self.set_result(model_parameters)

        return r2

    @classmethod
    def transform_input(cls, **kwargs):
        """
        Transform and validate input kwargs.

        :param kwargs: Dict;
        :return: Dict;
        """
        return cls.TRANSFORM_PROPERTIES_CLS.transform_input(**kwargs)

    def init_properties(self, **kwargs):
        """
        Setup system properties from input.

        :param kwargs: Dict; all supplied input properties
        """
        logger.debug(f"initialising properties of analytics task {self.name}")
        for kwarg in kwargs:
            setattr(self, kwarg, kwargs[kwarg])


class LCBinaryAnalyticsTask(AnalyticsTask):
    """
    Fitting task class aimed to fit light curves of eclipsing binary stars.
    """
    FIT_CLS = None
    PLOT_CLS = None
    FIT_PARAMS_COMBINATIONS = json.dumps({
        "standard": {"system": ["inclination", "eccentricity", "argument_of_periastron", "period",
                                "primary_minimum_time", "additional_light", "phase_shift"],
                     "primary": ["mass", "t_eff", "surface_potential", "gravity_darkening", "albedo",
                                 "synchronicity", "metallicity", "spots", "pulsations"],
                     "secondary": ["mass", "t_eff", "surface_potential", "gravity_darkening", "albedo",
                                   "synchronicity", "metallicity", "spots", "pulsations"],
                     "nuisance": ['ln_f']
                     },
        "community": {"system": ["inclination", "eccentricity", "argument_of_periastron", "period", "semi_major_axis",
                                 "primary_minimum_time", "additional_light", "phase_shift" "mass_ratio"],
                      "primary": ["t_eff", "surface_potential", "gravity_darkening", "albedo",
                                  "synchronicity", "metallicity", "spots", "pulsations"],
                      "secondary": ["t_eff", "surface_potential", "gravity_darkening", "albedo",
                                    "synchronicity", "metallicity", "spots", "pulsations"],
                      "nuisance": ['ln_f']
                      },
        "spots": ["longitude", "latituded", "angular_radius", "temperature_factor"],
        "pulsations": ["l", "m", "amplitude", "frequency", "start_phase", "mode_axis_theta", "mode_axis_phi"]
    }, indent=4)
    TRANSFORM_PROPERTIES_CLS = transform.LCBinaryAnalyticsProperties

    def __init__(self, method, expected_morphology='detached', name=None,
                 atmosphere_models=None, limb_darkening_coefficients=None, **kwargs):
        self.validate_method(method)
        if method in self.MCMC_NAMES:
            self.__class__.FIT_CLS = lambda: \
                lc_fit.LCFitMCMC(morphology=expected_morphology,
                                 atmosphere_model=atmosphere_models,
                                 limb_darkening_coefficients=limb_darkening_coefficients)
            self.__class__.PLOT_CLS = LCPlotMCMC
        elif method in self.LS_NAMES:
            self.__class__.FIT_CLS = lambda: \
                lc_fit.LCFitLeastSquares(morphology=expected_morphology,
                                         atmosphere_model=atmosphere_models,
                                         limb_darkening_coefficients=limb_darkening_coefficients)
            self.__class__.PLOT_CLS = LCPlotLsqr
        super().__init__(method, name, **kwargs)

    def load_result(self, filename, autofill_sma=True):
        return super().load_result(filename, autofill_sma=autofill_sma)

    def set_result(self, result, autofill_sma=True):
        return super().set_result(result, autofill_sma=autofill_sma)


class RVBinaryAnalyticsTask(AnalyticsTask):
    """
    Fitting task class aimed to fit radial velocity (RV) curves of eclipsing binary stars. For now, the method
    support only kinematic method for calculation of radial velocities (regarding stars as point masses).
    """
    FIT_CLS = None
    PLOT_CLS = None
    FIT_PARAMS_COMBINATIONS = json.dumps({

        'community': {
            'system': ['mass_ratio', 'asini', 'eccentricity', 'argument_of_periastron',
                       'gamma', 'period', 'primary_minimum_time'],
            'nuisance': ['ln_f']
        },

        'standard': {
            'primary': ['mass'],
            'secondary': ['mass'],
            'system': ['inclination', 'eccentricity', 'argument_of_periastron',
                       'gamma', 'period', 'primary_minimum_time'],
            'nuisance': ['ln_f']
        }
    }, indent=4)
    TRANSFORM_PROPERTIES_CLS = transform.RVBinaryAnalyticsTask

    def __init__(self, method, name=None, atmosphere_models=None, limb_darkening_factor=None, **kwargs):
        self.validate_method(method)
        if method in self.MCMC_NAMES:
            self.__class__.FIT_CLS = lambda: rv_fit.RVFitMCMC()
            self.__class__.PLOT_CLS = RVPlotMCMC
        elif method in self.LS_NAMES:
            self.__class__.FIT_CLS = lambda: rv_fit.RVFitLeastSquares()
            self.__class__.PLOT_CLS = RVPlotLsqr
        super().__init__(method, name, **kwargs)



import numpy as np


def surface_area_coverage(size, visible, visible_coverage, partial=None, partial_coverage=None):
    """
    Prepare array with coverage of surface areas.

    :param size: int; size of array
    :param visible: numpy.array; full visible areas (numpy fancy indexing), array like [False, True, True, False]
    :param visible_coverage: numpy.array; defines coverage of visible (coverage onTrue positions)
    :param partial: numpy.array; partial visible areas (numpy fancy indexing)
    :param partial_coverage: numpy.array; defines coverage of partial visible
    :return: numpy.array
    """
    # initialize zeros, since there is no input for invisible (it means everything what left after is invisible)
    coverage = np.zeros(size)
    coverage[visible] = visible_coverage
    if partial is not None:
        coverage[partial] = partial_coverage
    return coverage



import numpy as np

from abc import ABCMeta, abstractmethod
from copy import copy


class AbstractOrbit(metaclass=ABCMeta):
    """
     Abstract class that defines framework for (orbital) motion of the System instances.
    """
    MANDATORY_KWARGS = []
    OPTIONAL_KWARGS = []
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS

    def __init__(self, **kwargs):
        self.kwargs = copy(kwargs)

        self.period = np.nan
        self.inclination = np.nan
        self.phase_shift = 0.0

    @classmethod
    def true_phase(cls, phase, phase_shift):
        """
        Returns shifted phase of the orbit by the amount `phase_shift`.

        :param phase: Union[numpy.array, float];
        :param phase_shift: float;
        :return: Union[numpy.array, float];
        """
        return phase + phase_shift

    @staticmethod
    def phase(true_phase, phase_shift):
        """
        reverts the phase shift introduced in function true phase

        :param true_phase: Union[numpy.array, float];
        :param phase_shift: numpy.float;
        :return: Union[numpy.array, float];
        """
        return true_phase - phase_shift

    @abstractmethod
    def orbital_motion(self, phase):
        pass



import numpy as np
import matplotlib.gridspec as gridspec

from corner import corner as _corner
from matplotlib import pyplot as plt


class Plot(object):
    @staticmethod
    def corner(**kwargs):
        """
        Evaluate mcmc corner plot.

        """
        flat_chain = kwargs.pop('flat_chain')
        fit_params = kwargs.pop('fit_params')
        variable_labels = kwargs.pop('variable_labels')
        figure = _corner(flat_chain, **kwargs)

        # Extract the axes
        ndim = flat_chain.shape[1]
        axes = np.array(figure.axes).reshape((ndim, ndim))

        # Loop over the diagonal, adding units
        for i, label in enumerate(variable_labels):
            ax = axes[i, i]
            value = fit_params[label]['value']
            bottom = fit_params[label]["confidence_interval"]['min'] - value
            top = fit_params[label]["confidence_interval"]['max'] - value

            unit = fit_params[label]['unit']
            unit = '' if unit == 'dimensionless' or unit is None else unit
            if any(x in label for x in ['t_eff', 'argument_of_periastron']):
                title = r'{0}=${1:.0f}^{{{2:+.0f}}}_{{{3:+.0f}}}$ {4}'.format(kwargs['labels'][i], value, top, bottom,
                                                                              unit)
            elif any(x in label for x in ['eccentricity']):
                title = r'{0}=${1:.3f}^{{{2:+.3f}}}_{{{3:+.3f}}}$ {4}'.format(kwargs['labels'][i], value, top, bottom,
                                                                              unit)
            else:
                title = r'{0}=${1:.3f}^{{{2:+.3f}}}_{{{3:+.3f}}}$ {4}'.format(kwargs['labels'][i], value, top, bottom,
                                                                              unit)
            ax.set_title(title)

        plt.show()

    @staticmethod
    def paramtrace(**kwargs):
        """
        Show traces of mcmc chain.
        """
        labels = kwargs['labels']
        hash_map = {label: idx for idx, label in enumerate(kwargs['variable_labels']) 
                    if label in kwargs['traces_to_plot']}

        height = len(kwargs['traces_to_plot'])
        fig = plt.figure(figsize=(8, 2.5*height))

        gs = gridspec.GridSpec(height, 1)
        ax = []

        plot_conter = 0
        for idx, label in enumerate(kwargs['variable_labels']):
            if label not in kwargs['traces_to_plot']:
                continue
            ax.append(fig.add_subplot(gs[plot_conter])) \
                if plot_conter == 0 else ax.append(fig.add_subplot(gs[plot_conter], sharex=ax[0]))
            plot_conter += 1
            ax[-1].scatter(np.arange(kwargs['flat_chain'].shape[0]), kwargs['flat_chain'][:, hash_map[label]],
                           label=labels[idx], s=0.2)
            ax[-1].legend(loc=1)

            unit = kwargs['fit_params'][label]['unit']
            unit = '' if unit == 'dimensionless' or unit is None else ' / [{0}]'.format(unit)
            ax[-1].set_ylabel(f'{labels[idx]}{unit}')

            if kwargs['truths']:
                ax[-1].axhline(kwargs['fit_params'][label]['value'], linestyle='dashed', color='black')
                ax[-1].axhline(kwargs['fit_params'][label]["confidence_interval"]['min'],
                               linestyle='dotted', color='black')
                ax[-1].axhline(kwargs['fit_params'][label]["confidence_interval"]['max'],
                               linestyle='dotted', color='black')

        ax[-1].set_xlabel('N')

        plt.subplots_adjust(right=1.0, top=1.0, hspace=0)
        plt.show()

    @staticmethod
    def autocorr(**kwargs):
        """
        Show autocorrelation function.
        """
        labels = kwargs['labels']
        hash_map = {label: idx for idx, label in enumerate(kwargs['variable_labels']) if label in
                    kwargs['correlations_to_plot']}

        height = len(kwargs['correlations_to_plot'])
        fig = plt.figure(figsize=(8, 2.5 * height))

        gs = gridspec.GridSpec(height, 1)
        ax = []

        for idx, label in enumerate(kwargs['variable_labels']):
            if label not in kwargs['correlations_to_plot']:
                continue

            ax.append(fig.add_subplot(gs[idx])) if idx == 0 else ax.append(fig.add_subplot(gs[idx], sharex=ax[0]))

            lbl = 'corr_time = {0:.2f}'.format(kwargs['autocorr_time'][idx])
            ax[-1].scatter(np.arange(kwargs['autocorr_fns'].shape[0]), kwargs['autocorr_fns'][:, hash_map[label]],
                           label=lbl, s=0.2)
            ax[-1].set_ylabel(f'{labels[idx]} correlation fn')
            ax[-1].legend()

        ax[-1].set_xlabel('N')

        plt.subplots_adjust(right=1.0, top=1.0, hspace=0)
        plt.show()



import numpy as np

from ... binary_system import t_layer


def lightcurves_mean_error(lc):
    """
    If observation errors are not provided, the default 5 percent relative error
    is used to generate synthetic errors.

    :param lc: numpy.array; light curve
    :return: numpy.array; synthetic errors
    """
    return np.mean(lc) * 0.05


def radialcurves_mean_error(rv):
    """
    If observation errors are not provided, the default 5 percent relative error
    is used to generate synthetic errors.

    :param rv: numpy.array; radial velocities
    :return: numpy.array; synthetic errors
    """
    return np.mean(rv) * 0.05


def is_time_dependent(labels):
    """
    If 'system@primary_minimum_time' is located in the fit parameters, the fit parameters are considered
    time dependent and observations are therefore expected to be supplied in JD.

    :param labels: List[str]; parameter labels
    :return: bool
    """
    if 'system@period' in labels and 'system@primary_minimum_time' in labels:
        return True
    return False


def time_layer_resolver(x_data, pop=False, **kwargs):
    """
    If kwargs contain `period` and `primary_minimum_time`, then xs is expected to be JD
    time not phases. Then, x_data (observational time) has to be converted to phases.

    :param pop: bool; determine if remove the system@primary_minimum_time parameter from the fit
                      parameters or just read it
    :param x_data: Union[List, numpy.array];
    :param kwargs: Dict;
    :return: Tuple;
    """

    if is_time_dependent(list(kwargs.keys())):
        t0 = kwargs['system@primary_minimum_time']
        if pop:
            kwargs.pop('system@primary_minimum_time')
        period = kwargs['system@period']
        x_data_new = t_layer.jd_to_phase(t0, period, x_data, centre=0.5)
    else:
        x_data_new = x_data % 1.0
    return x_data_new, kwargs



import numpy as np

from ... pypex.base import shape
from ... pypex.poly2d.intersection import linter
from ... pypex.base.conf import ROUND_PRECISION


class Line(shape.Shape2D):
    __intersect__ = ['INTERSECT']
    __overlapping__ = ['OVERLAP']

    def __str__(self):
        return "Line: [{}]".format(", ".join([str(v) for v in self.hull]))

    def __repr__(self):
        return "Line: [{}]".format(", ".join([str(v) for v in self.hull]))

    def intersects(self, line, _full=False, in_touch=False, round_tol=ROUND_PRECISION):
        """
        Figure out whether two line are in intersection or not

        :param round_tol: int; consider as same up to 'round_tol' decimal numbers
        :param in_touch: bool
        :param line: pypex.poly2d.line.Line
        :param _full: bool; define whether return full output or not
        :return: bool or tuple
        """

        # fixme: return dual type is probably not a good idea
        intersection = linter.intersection(self.hull[0], self.hull[1], line.hull[0], line.hull[1],
                                           in_touch=in_touch, round_tol=round_tol)
        if _full:
            return intersection
        return intersection[1] and (intersection[4] in "INTERSECT")

    def full_intersects(self, line, in_touch=False, round_tol=ROUND_PRECISION):
        """
        Figure out whether two line are in intersection or not.
        Method is here to avoid repeating evaluation of if condition statement in case of _full=True in intersects()
        method in case when is necessary to find intersection of huge amount of lines in loop

        :param round_tol: int; consider as same up to 'round_tol' decimal numbers
        :param in_touch: bool
        :param line: pypex.poly2d.line.Line
        :return: tuple
        """
        return linter.intersection(self.hull[0], self.hull[1], line.hull[0], line.hull[1],
                                   in_touch=in_touch, round_tol=round_tol)

    def intersection(self, line, in_touch=False, round_tol=ROUND_PRECISION):
        """
        Find intersection point of two lines if exists.

        :param round_tol: int
        :param in_touch: bool
        :param line: pypex.poly2d.line.Line
        :return: pypex.poly2d.point.Point or None
        """

        intersection = self.full_intersects(line, in_touch=in_touch, round_tol=round_tol)
        intersect = intersection[1] and (intersection[4] in "INTERSECT")
        if not intersect:
            return
        return intersection[2]

    def to_array(self):
        """
        Get points of line in numpy array.
        :return: numpy.array;
        """
        return np.array([point.to_array() for point in self.to_Points()])

    def sort_clockwise(self, *args, **kwargs):
        return self.hull

    def direction_vector(self):
        """
        Get direction vector.
        :return: numpy.array;
        """
        return self.hull[1] - self.hull[0]

    def parametrized(self):
        """
        Return callable parametrization of given line as function.
        """
        def _parametrized(t):
            v = self.direction_vector()
            return self.hull[0][0] + (t * v[0]), self.hull[0][1] + (t * v[1])
        return _parametrized

    def angle(self, other: shape.Shape2D, degrees=False):
        """
        Return angle between vectors defined by self `Line` and other 'Line'.
        """
        vector_self = self.hull[0] - self.hull[1]
        vector_other = other.hull[0] - other.hull[1]

        unit_vector_self = vector_self / np.linalg.norm(vector_self)
        unit_vector_other = vector_other / np.linalg.norm(vector_other)
        dot_product = np.dot(unit_vector_self, unit_vector_other)
        angle = np.arccos(dot_product)
        if degrees:
            angle = np.degrees(angle)
        return angle



import numpy as np

from elisa.logger import getLogger

from ... const import Position
from ... graphic import graphics
from .. container import SinglePositionContainer
from .. import utils as sutils
from ... base.graphics import plot
from .. curves import utils as crv_utils
from ... observer.observer import Observer


logger = getLogger('single_system.graphic.animation')


class Animation(object):
    defpos = Position(*(0, np.nan, 0.0, np.nan, 0.0))

    def __init__(self, instance):
        self.single = instance

    def rotational_motion(self, start_phase=-0.5, stop_phase=0.5, phase_step=0.01, scale='linear',
                          colormap=None, savepath=None, plot_axis=True, subtract_equilibrium=False, edges=False):
        """
        Function creates animation of the rotational motion.

        :param start_phase: float; starting phase of the animation
        :param stop_phase: float; end phase of the animation
        :param phase_step: float; phase step between animation frames
        :param scale: str; `linear` or `log`, scale of the colormap
        :param colormap: str;
        :param savepath: str; animation will be stored to `savepath`
        :param subtract_equilibrium: bool; equilibrium part of the quantity is removed (for pulsations)
        :param plot_axis: bool; if False, axis will be hidden
        :param edges: bool; highlight edges of surface faces
        :**available colormap options**:
            * :'gravity_acceleration': surface distribution of gravity acceleration,
            * :'temperature': surface distribution of the effective temperature,
            * :'velocity': absolute values of surface elements velocities with respect to the observer,
            * :'radial_velocity': radial component of the surface element velocities relative to the observer,
            * :'normal_radiance': surface element radiance perpendicular to the surface element,
            * :'radiance': radiance of the surface element in a direction towards the observer,
            * :'radius': distance of the surface elements from the centre of mass
            * :'horizontal_displacement': distribution of the horizontal component of surface displacement
            * :'horizontal_acceleration': distribution of horizontal component surface acceleration
            * :'v_r_perturbed': radial component of the pulsation velocity (perpendicular towards
            * :'v_horizontal_perturbed': horizontal component of the pulsation  velocity
        """
        anim_kwargs = dict()

        if stop_phase < start_phase:
            raise ValueError(f'Starting phase {start_phase} is greater than stop phase {stop_phase}')

        n_frames = int((stop_phase - start_phase) / phase_step)
        phases = np.linspace(start_phase, stop_phase, num=n_frames)
        points = []
        faces = []
        cmap = []

        orbital_motion = \
            self.single.calculate_lines_of_sight(input_argument=phases, return_nparray=False, calculate_from='phase')

        # calculating radiances
        o = Observer(passband=['bolometric', ], system=self.single)
        atm_kwargs = dict(
            passband=o.passband,
            left_bandwidth=o.left_bandwidth,
            right_bandwidth=o.right_bandwidth,
        )

        logger.info('calculating surface parameters (points, faces, colormap)')
        for pos_idx, position in enumerate(orbital_motion):
            from_this = dict(single_system=self.single, position=self.defpos)
            on_pos = SinglePositionContainer.from_single_system(**from_this)
            on_pos.set_on_position_params(position)
            on_pos.set_time()
            on_pos.build()

            crv_utils.prep_surface_params(
                system=on_pos, write_to_containers=True, **atm_kwargs
            )

            on_pos = sutils.move_sys_onpos(on_pos, position, on_copy=False)

            star = getattr(on_pos, 'star')
            points.append(star.points)
            faces.append(star.faces)

            args = (colormap, star, position.phase, 0.0, 1.0, self.single.inclination, on_pos.position)
            kwargs = dict(scale=scale, unit='default', subtract_equilibrium=subtract_equilibrium)
            cmap.append(plot.add_colormap_to_plt_kwargs(*args, **kwargs))

        anim_kwargs.update({
            'start_phase': start_phase,
            'stop_phase': stop_phase,
            'n_frames': n_frames,
            'phases': phases,
            'points': points,
            'faces': faces,
            'cmap': cmap,
            'axis_lim': 1.2 * np.max(points[0]),
            'savepath': savepath,
            'colormap': colormap,
            "plot_axis": plot_axis,
            'edges': edges
        })
        logger.debug('Passing parameters to graphics module')
        graphics.single_surface_anim(**anim_kwargs)





class Buffer(object):
    """
    Buffers for parameters loaded from external files in singleton format
    """
    _instance = None

    MAX_STORAGE = 300

    LD_CFS_TABLES = dict()
    ATMOSPHERE_TABLES = dict()

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Buffer, cls).__new__(cls)
            cls.DEFAULT_SETTINGS = cls.settings_serializer()
        return cls._instance

    @classmethod
    def settings_serializer(cls):
        return {
            "LD_CFS_TABLES": cls.LD_CFS_TABLES,
            "ATMOSPHERE_TABLES": cls.ATMOSPHERE_TABLES,
            "MAX_STORAGE": cls.MAX_STORAGE
        }

    @classmethod
    def reduce_buffer(cls, storage):
        """
        If buffer exceeds allowed size, the first items are deleted.

        :param storage: Dict; buffer
        :return: Dict; reduced buffer
        """
        if len(storage) > cls.MAX_STORAGE:
            start_idx = len(storage)-cls.MAX_STORAGE
            for key in list(storage.keys())[:start_idx]:
                del storage[key]
        return storage


buffer = Buffer()



import numpy as np

from .. container import SinglePositionContainer
from ... import units as u
from ... base import transform
from ... base.graphics import plot
from ... const import Position
from ... graphic import graphics
from ... base.surface.faces import correct_face_orientation
from .. import utils as sutils
from .. curves import utils as crv_utils
from ... observer.observer import Observer
from ... import umpy as up


class Plot(object):
    """
    Universal plot interface for binary system class, more detailed documentation for each value of descriptor is
    available in graphics library::

        `equipotential` - plots crossection of surface Hill planes in xz plane
        `wireframe` - wire frame model of the star
        `mesh` - plot surface points
        `surface` - plot stellar surfaces
    """

    defpos = Position(*(0, np.nan, 0.0, np.nan, 0.0))

    def __init__(self, instance):
        self.single = instance

    def equipotential(self, axis_unit=u.solRad, return_figure_instance=False):
        """
        Function for quick 2D plot of equipotential cross-section in xz plane.

        :param axis_unit: Union[astropy.unit, dimensionless]; - axis units
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        """
        equipotential_kwargs = dict()

        points = self.single.calculate_equipotential_boundary()
        points = (points * u.DefaultSingleSystemUnits.star.equivalent_radius).to(axis_unit)

        equipotential_kwargs.update({
            'return_figure_instance': return_figure_instance,
            'points': points,
            'axis_unit': axis_unit,
        })
        return graphics.equipotential_single_star(**equipotential_kwargs)

    def mesh(self, phase=0.0, plot_axis=True, axis_unit=u.solRad, inclination=None, azimuth=None,
             return_figure_instance=False):
        """
        Function plots 3D scatter plot of the surface points.

        :param phase: float;
        :param plot_axis: bool; switch the plot axis on/off
        :param axis_unit: Union[astropy.unit, dimensionless]; - axis units
        :param inclination: Union[float, astropy.Quantity]; in degree - elevation of camera
        :param azimuth: Union[float, astropy.Quantity]; camera azimuth
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        """
        single_mesh_kwargs = dict()

        inclination = transform.deg_transform(inclination, u.deg, transform.WHEN_FLOAT64,
                                              u.DefaultSingleSystemInputUnits.system.inclination) \
            if inclination is not None else up.degrees(self.single.inclination)
        azim = self.single.orbit.rotational_motion(phase=phase)[0][0]
        azimuth = transform.deg_transform(azimuth, u.deg, when_float64=transform.WHEN_FLOAT64) \
            if azimuth is not None else up.degrees(azim) - 90

        position_container = SinglePositionContainer.from_single_system(self.single, self.defpos)
        position_container.build_mesh()
        position_container.build_perturbations()

        mesh = position_container.star.get_flatten_parameter('points')
        denominator = (1 * axis_unit.to(u.DefaultSingleSystemUnits.star.equivalent_radius))
        mesh /= denominator
        equatorial_radius = \
            position_container.star.equatorial_radius * u.DefaultSingleSystemUnits.star.equivalent_radius.to(axis_unit)

        single_mesh_kwargs.update({
            'return_figure_instance': return_figure_instance,
            'phase': phase,
            'axis_unit': axis_unit,
            'plot_axis': plot_axis,
            "inclination": inclination,
            "azimuth": azimuth,
            "mesh": mesh,
            'equatorial_radius': equatorial_radius,
        })

        return graphics.single_star_mesh(**single_mesh_kwargs)

    def wireframe(self, phase=0.0, plot_axis=True, axis_unit=u.solRad, inclination=None, azimuth=None,
                  return_figure_instance=False):
        """
        Returns 3D wireframe of the object.

        :param phase: float;
        :param plot_axis: bool; switch the plot axis on/off
        :param axis_unit: Union[astropy.unit, dimensionless]; - axis units
        :param inclination: Union[float, astropy.Quantity]; in degree - elevation of camera
        :param azimuth: Union[float, astropy.Quantity]; camera azimuth
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        """
        wireframe_kwargs = dict()

        inclination = transform.deg_transform(inclination, u.deg, transform.WHEN_FLOAT64,
                                              u.DefaultSingleSystemInputUnits.system.inclination) \
            if inclination is not None else up.degrees(self.single.inclination)
        azim = self.single.orbit.rotational_motion(phase=phase)[0][0]
        azimuth = transform.deg_transform(azimuth, u.deg, when_float64=transform.WHEN_FLOAT64) \
            if azimuth is not None else up.degrees(azim) - 90

        position_container = SinglePositionContainer.from_single_system(self.single, self.defpos)
        position_container.build_mesh()
        position_container.build_faces()

        points, faces = position_container.star.surface_serializer()
        denominator = (1 * axis_unit.to(u.DefaultSingleSystemUnits.star.equivalent_radius))
        points /= denominator
        equatorial_radius = \
            position_container.star.equatorial_radius * u.DefaultSingleSystemUnits.star.equivalent_radius.to(axis_unit)

        wireframe_kwargs.update({
            'return_figure_instance': return_figure_instance,
            'phase': phase,
            'axis_unit': axis_unit,
            'plot_axis': plot_axis,
            "inclination": inclination,
            "azimuth": azimuth,
            "mesh": points,
            "triangles": faces,
            'equatorial_radius': equatorial_radius,
        })

        return graphics.single_star_wireframe(**wireframe_kwargs)

    def surface(self, phase=0.0, normals=False, edges=False, colormap=None, plot_axis=True, face_mask=None,
                elevation=None, azimuth=None, colorbar_unit='default', axis_unit=u.solRad,
                colorbar_orientation='vertical', colorbar=True, scale='linear', surface_color='g',
                colorbar_separation=0.0, colorbar_size=0.7, return_figure_instance: bool=False,
                subtract_equilibrium: bool=False):
        """
        Function creates plot of single system components.

        :param phase: float; phase at which plot the system, important for eccentric orbits
        :param normals: bool; plot normals of the surface phases as arrows
        :param edges: bool; highlight edges of surface faces
        :param colormap: str;
        :param plot_axis: bool; if False, axis will be hidden
        :param face_mask: array[bool]; mask to select which faces to display
        :param elevation: Union[float, astropy.Quantity]; in degree - elevation of camera
        :param azimuth: Union[float, astropy.Quantity]; camera azimuth
        :param colorbar_unit: str; colormap unit
        :param axis_unit: Union[astropy.unit, dimensionless]; - axis units
        :param colorbar_orientation: `horizontal` or `vertical` (default)
        :param colorbar: bool; colorbar on/off switch
        :param scale: str; `linear` or `log`
        :param surface_color: Tuple; tuple of colors for components if `colormap` is not specified
        :param colorbar_separation: float; shifting position of the colorbar from its default postition, default is 0.0
        :param colorbar_size: float; relative size of the colorbar, default 0.7
        :param return_figure_instance: bool; if True, the Figure instance is returned instead of displaying the
                                             produced figure
        :param subtract_equilibrium: bool; if True, equilibrium values are subtracted from the colormap
        :**available colormap options**:
            * :'gravity_acceleration': surface distribution of gravity acceleration,
            * :'temperature': surface distribution of the effective temperature,
            * :'velocity': absolute values of surface elements velocities with respect to the observer,
            * :'radial_velocity': radial component of the surface element velocities relative to the observer,
            * :'normal_radiance': surface element radiance perpendicular to the surface element,
            * :'radiance': radiance of the surface element in a direction towards the observer,
            * :'radius': distance of the surface elements from the centre of mass
            * :'horizontal_displacement': distribution of the horizontal component of surface displacement
            * :'horizontal_acceleration': distribution of horizontal component surface acceleration
            * :'v_r_perturbed': radial component of the pulsation velocity (perpendicular towards
            * :'v_horizontal_perturbed': horizontal component of the pulsation  velocity

        """
        surface_kwargs = dict()

        elevation = transform.deg_transform(elevation, u.deg, when_float64=transform.WHEN_FLOAT64) \
            if elevation is not None else 0
        azimuth = transform.deg_transform(azimuth, u.deg, when_float64=transform.WHEN_FLOAT64) \
            if azimuth is not None else 180

        single_position = self.single.orbit.rotational_motion(phase=phase)[0]
        single_position = Position(0, np.nan, single_position[0], single_position[1], single_position[2])

        position_container = SinglePositionContainer.from_single_system(self.single, self.defpos)
        position_container.set_on_position_params(single_position)
        position_container.set_time()
        position_container.build(phase=phase, build_pulsations=True)
        correct_face_orientation(position_container.star, com=0)

        # calculating radiances
        o = Observer(passband=['bolometric', ], system=self.single)
        atm_kwargs = dict(
            passband=o.passband,
            left_bandwidth=o.left_bandwidth,
            right_bandwidth=o.right_bandwidth,
        )

        crv_utils.prep_surface_params(
            system=position_container, write_to_containers=True, **atm_kwargs
        )

        position_container = sutils.move_sys_onpos(position_container, single_position)

        star_container = getattr(position_container, 'star')

        args = (colormap, star_container, phase, 0.0, 1.0, self.single.inclination, position_container.position)
        kwargs = dict(scale=scale, unit=colorbar_unit, subtract_equilibrium=subtract_equilibrium)
        surface_kwargs.update({'cmap': plot.add_colormap_to_plt_kwargs(*args, **kwargs)})

        surface_kwargs.update({
            'points': star_container.points,
            'triangles': star_container.faces
        })

        face_mask = np.ones(star_container.faces.shape[0], dtype=bool) if face_mask is None else face_mask
        surface_kwargs['triangles'] = surface_kwargs['triangles'][face_mask]
        if 'colormap' in surface_kwargs.keys():
            surface_kwargs['cmap'] = surface_kwargs['cmap'][face_mask]

        if normals:
            face_centres = star_container.get_flatten_parameter('face_centres')
            norm = star_container.get_flatten_parameter('normals')
            surface_kwargs.update({
                'centres': face_centres[face_mask],
                'arrows': norm[face_mask]
            })

        # normals
        unit_mult = (1*u.DefaultSingleSystemUnits.star.equivalent_radius).to(axis_unit).value
        surface_kwargs['points'] *= unit_mult

        if normals:
            surface_kwargs['centres'] *= unit_mult

        equatorial_radius = (star_container.equatorial_radius*u.DefaultSingleSystemUnits.star.equivalent_radius).\
            to(axis_unit).value

        surface_kwargs.update({
            'phase': phase,
            'normals': normals,
            'edges': edges,
            'colormap': colormap,
            'plot_axis': plot_axis,
            'face_mask': face_mask,
            "elevation": elevation,
            "azimuth": azimuth,
            'unit': colorbar_unit,
            'axis_unit': axis_unit,
            'colorbar_orientation': colorbar_orientation,
            'colorbar': colorbar,
            'scale': scale,
            'equatorial_radius': equatorial_radius,
            'surface_color': surface_color,
            'colorbar_separation': colorbar_separation,
            'colorbar_size': colorbar_size,
            'return_figure_instance': return_figure_instance
        })
        return graphics.single_star_surface(**surface_kwargs)



# DO NOT CHANGE THIS!!!
SUN = {
    "system": {
        "inclination": "90 deg",
        "rotation_period": "25.38 d",
        "gamma": "0 m/s",
        "reference_time": 0.0,
        "phase_shift": 0.0,
        "distance": "10 pc"
    },
    "star": {
        "mass": "1.0 solMass",
        "t_eff": "5772 K",
        "metallicity": 0.0,
        "equivalent_radius": "1.0 solRad"
    }
}



import numpy as np

from copy import copy
from .. container import OrbitalPositionContainer
from .. import utils as butils, dynamic
from ... const import Position
from ... logger import getLogger
from ... graphic import graphics
from ... base.graphics import plot
from .. curves import utils as crv_utils
from ... observer.observer import Observer

logger = getLogger('binary_system.graphic.animation')


class Animation(object):
    defpos = Position(*(0, 1.0, 0.0, 0.0, 0.0))

    def __init__(self, instance):
        self.binary = instance

    def orbital_motion(self, start_phase=-0.5, stop_phase=0.5, phase_step=0.01, units='cgs', scale='linear',
                       colormap=None, savepath=None, separate_colormaps=None, subtract_equilibrium=False,
                       plot_axis=True, edges=False):
        """
        Function creates animation of the orbital motion.

        :param start_phase: float; starting phase of the animation
        :param stop_phase: float; end phase of the animation
        :param phase_step: float; phase step between animation frames
        :param units: str; unit type of surface colormap `SI` or `cgs`
        :param scale: str; `linear` or `log`, scale of the colormap
        :param colormap: str;
        :param savepath: str; animation will be stored to `savepath`
        :param separate_colormaps: bool; if True, figure will contain separate colormap for each component
        :param subtract_equilibrium: bool; equilibrium part of the quantity is removed (for pulsations)
        :param plot_axis: bool; if False, axis will be hidden
        :param edges: bool; highlight edges of surface faces
        :**available colormap options**:
            * :'gravity_acceleration': surface distribution of gravity acceleration,
            * :'temperature': surface distribution of the effective temperature,
            * :'velocity': absolute values of surface elements velocities with respect to the observer,
            * :'radial_velocity': radial component of the surface element velocities relative to the observer,
            * :'normal_radiance': surface element radiance perpendicular to the surface element,
            * :'radiance': radiance of the surface element in a direction towards the observer,
            * :'radius': distance of the surface elements from the centre of mass
            * :'horizontal_displacement': distribution of the horizontal component of surface displacement
            * :'horizontal_acceleration': distribution of horizontal component surface acceleration
            * :'v_r_perturbed': radial component of the pulsation velocity (perpendicular towards
            * :'v_horizontal_perturbed': horizontal component of the pulsation  velocity

        """
        anim_kwargs = dict()

        if stop_phase < start_phase:
            raise ValueError(f'Starting phase {start_phase} is greater than stop phase {stop_phase}')

        components = butils.component_to_list('both')

        if separate_colormaps is None:
            separate_colormaps = self.binary.morphology != 'over-contact' and \
                                 colormap not in ['velocity', 'radial_velocity']

        n_frames = int((stop_phase - start_phase) / phase_step)
        phases = np.linspace(start_phase, stop_phase, num=n_frames)
        none = [None for _ in range(n_frames)]
        points = {component: copy(none) for component in components}
        faces = {component: copy(none) for component in components}
        cmap = {component: copy(none) for component in components}

        orbital_motion = \
            self.binary.calculate_orbital_motion(input_argument=phases, return_nparray=False, calculate_from='phase')

        # in case of asynchronous component rotation and spots, the positions of spots are recalculated
        spots_longitudes = dynamic.calculate_spot_longitudes(self.binary, phases, component="all")
        potentials = self.binary.correct_potentials(phases, component="all", iterations=2)
        distances_to_com = {'primary': np.nan, 'secondary': np.nan}

        logger.info('calculating surface parameters (points, faces, colormap)')
        for pos_idx, position in enumerate(orbital_motion):
            from_this = dict(binary_system=self.binary, position=self.defpos)
            on_pos = OrbitalPositionContainer.from_binary_system(**from_this)
            on_pos.time = 86400 * self.binary.period * position.phase
            dynamic.assign_spot_longitudes(on_pos, spots_longitudes, index=pos_idx, component="all")
            on_pos.set_on_position_params(position, potentials["primary"][pos_idx],
                                          potentials["secondary"][pos_idx])
            on_pos.build(components_distance=position.distance)

            # calculating radiances
            o = Observer(passband=['bolometric', ], system=self.binary)
            atm_kwargs = dict(
                passband=o.passband,
                left_bandwidth=o.left_bandwidth,
                right_bandwidth=o.right_bandwidth,
            )

            crv_utils.prep_surface_params(
                system=on_pos, write_to_containers=True, **atm_kwargs
            )

            com = {'primary': 0.0, 'secondary': position.distance}
            distances_to_com = {'primary': position.distance * self.binary.mass_ratio / (1 + self.binary.mass_ratio),
                                'secondary': position.distance / (1 + self.binary.mass_ratio)}
            on_pos.primary.points[:, 0] -= distances_to_com['primary']
            on_pos.secondary.points[:, 0] -= distances_to_com['primary']

            on_pos = butils.move_sys_onpos(on_pos, position, potentials["primary"][pos_idx],
                                           potentials["secondary"][pos_idx], on_copy=False)

            for component in components:
                star = getattr(on_pos, component)
                points[component][pos_idx], faces[component][pos_idx] = star.points, star.faces

                args = (colormap, star, position.phase, com[component], self.binary.semi_major_axis,
                        self.binary.inclination, on_pos.position)
                kwargs = dict(scale=scale, unit='default', subtract_equilibrium=subtract_equilibrium)
                cmap[component][pos_idx] = plot.add_colormap_to_plt_kwargs(*args, **kwargs)

        anim_kwargs.update({
            'morphology': self.binary.morphology,
            'start_phase': start_phase,
            'stop_phase': stop_phase,
            'n_frames': n_frames,
            'phases': phases,
            'points_primary': points['primary'],
            'points_secondary': points['secondary'],
            'faces_primary': faces['primary'],
            'faces_secondary': faces['secondary'],
            'primary_cmap': cmap['primary'],
            'secondary_cmap': cmap['secondary'],
            'axis_lim': 1.3*np.max((distances_to_com['primary'], distances_to_com['secondary'])),
            'savepath': savepath,
            'colormap': colormap,
            "separate_colormaps": separate_colormaps,
            "plot_axis": plot_axis,
            "edges": edges
        })
        logger.debug('Passing parameters to graphics module')
        graphics.binary_surface_anim(**anim_kwargs)



"""
implementation of SAP algorithm in 2d
detection of collision of two faces
if will also return false if faces are just in touch

overlap is handled as no intersection at all
"""
import numpy as np

from .... pypex.poly2d import projection
from .... pypex.base.conf import ROUND_PRECISION


def separating_axis_theorem(poly1, poly2, in_touch=False, round_tol=ROUND_PRECISION):
    """

    :param round_tol: int
    :param in_touch: bool
    :param poly1: numpy.array;
    :param poly2: numpy.array
    :return: bool
    """
    # number of vertices in each polygon
    nv1, nv2 = len(poly1), len(poly2)

    if (nv1 < 3) or (nv2 < 3):
        raise ValueError("invalid polygon shape, supported at least polygon with 3 corners")

    edges = np.concatenate(([[poly1[i], poly1[i + 1]] for i in range(-1, nv1 - 1)],
                            [[poly2[i], poly2[i + 1]] for i in range(-1, nv2 - 1)]), axis=0)
    for edge in edges:
        # tangent vector of testing line
        tangent = (edge[1] - edge[0]) / np.linalg.norm(edge[1] - edge[0])
        # normal vector of testing line
        normal = -tangent[1], tangent[0]

        # we flip tangent and normal vector (all vertices of all polygons will be projected on line defined by normal
        # vector isntead of tangential); just convinience
        tangent, normal = normal, tangent

        # projection of each vertex on testing line (basically it is going to transform coo of each vertex
        # to coordinate system, where tangential vector represents new `x` axis)

        projection_poly1 = np.array([projection.cartesian_to_vectors_defined(
            tangent, normal, projection.projection(vertex, tangent)) for vertex in poly1])
        projection_poly2 = np.array([projection.cartesian_to_vectors_defined(
            tangent, normal, projection.projection(vertex, tangent)) for vertex in poly2])

        # new x coo (y should and has to be zero now)
        projection_poly1_x, projection_poly2_x = projection_poly1.T[0], projection_poly2.T[0]

        # maximal length projected of each face
        projection_edge1, projection_edge2 = [round(projection_poly1_x.min(), round_tol),
                                              round(projection_poly1_x.max(), round_tol)], \
                                             [round(projection_poly2_x.min(), round_tol),
                                              round(projection_poly2_x.max(), round_tol)]

        projection_edge = [projection_edge1, projection_edge2]
        projection_edge.sort(key=lambda x: x[0])

        # if intervals connected in point or separated return True
        eval_method = np.less if in_touch else np.less_equal
        if eval_method(projection_edge[0][1], projection_edge[1][0]):
            return True
    return False


def separating_axis_theorem_line_adapt(line1, line2, in_touch=False, round_tol=ROUND_PRECISION):
    """
    test overlap of segments that are situated on the same line

    :param round_tol: int
    :param in_touch: bool
    :param line1: numpy.array
    :param line2: numpy.array
    :return: bool
    """
    tangent = (line1[1] - line1[0]) / np.linalg.norm(line1[1] - line1[0])
    normal = -tangent[1], tangent[0]

    projection_line1 = np.array([projection.cartesian_to_vectors_defined(
        tangent, normal, projection.projection(vertex, tangent)) for vertex in line1])
    projection_line2 = np.array([projection.cartesian_to_vectors_defined(
        tangent, normal, projection.projection(vertex, tangent)) for vertex in line2])
    projection_x = [np.round(projection_line1.T[0], round_tol), np.round(projection_line2.T[0], round_tol)]
    projection_x.sort(key=lambda x: x[0])

    eval_method = np.less if in_touch else np.less_equal
    if eval_method(projection_x[0][1], projection_x[1][0]):
        return True
    return False


def intersects(poly1, poly2, in_touch=False, round_tol=ROUND_PRECISION):
    """
    Resolve whether two objects intersetcs.

    In case of two segments it will resolve correctly only segments situated on tme same line.
    It is suppose to resolve whether these two lines are separated or overlapped/touched in point.

    :param round_tol: int; consider two numbers as same if match up to `round_tol` decimal numbers
    :param in_touch: bool
    :param poly1: numpy.array; convex polygon
    :param poly2: numpy.array; convex polygon
    :return: bool
    """
    if (len(poly1) == 2) and (len(poly2) == 2):
        return not separating_axis_theorem_line_adapt(poly1, poly2, in_touch, round_tol)
    return not separating_axis_theorem(poly1, poly2, in_touch, round_tol)



from ... import atm
from ... import settings
from ... observer.passband import init_bolometric_passband
from ... base.curves.utils import get_component_limbdarkening_cfs


def prep_surface_params(system, return_values=True, write_to_containers=False, **kwargs):
    """
    Prepares normal radiance and limb darkening coefficients variables.

    :param system: elisa.single_system.container.SystemContainer;
    :param return_values: bool; return normal radiances and limb darkening coefficients
    :param write_to_containers: bool; calculated values will be assigned to `system` container
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float

    :return:
    """
    # obtain limb darkening factor for each face
    ld_cfs = get_limbdarkening_cfs(system, **kwargs)

    # compute normal radiance for each face and each component
    normal_radiance = get_normal_radiance(system, **kwargs)

    # checking if `bolometric`filter is already used
    if 'bolometric' in ld_cfs['star'].keys():
        bol_ld_cfs = {'star': {'bolometric': ld_cfs['star']['bolometric']}}
    else:
        passband, left_bandwidth, right_bandwidth = init_bolometric_passband()
        bol_kwargs = {
            'passband': {'bolometric': passband},
            'left_bandwidth': left_bandwidth,
            'right_bandwith': right_bandwidth,
            'atlas': 'whatever'
        }
        bol_ld_cfs = get_limbdarkening_cfs(system, **bol_kwargs)

    normal_radiance = atm.correct_normal_radiance_to_optical_depth(normal_radiance, bol_ld_cfs)

    if write_to_containers:
        star = getattr(system, 'star')
        setattr(star, 'normal_radiance', normal_radiance['star'])
        setattr(star, 'ld_cfs', ld_cfs['star'])

    return normal_radiance, ld_cfs if return_values else None


def get_normal_radiance(system, **kwargs):
    """
    Compute normal radiance for all faces in SingleOrbitalPositionContainer.

    :param system: elisa.single_system.container.SystemContainer;
    :param kwargs: Dict; arguments to be passed into light curve generator functions
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float

    :return: Dict[String, numpy.array]
    """
    star = system.star
    symmetry_test = not system.has_spots() and not system.has_pulsations()

    # utilizing surface symmetry in case of a clear surface
    if symmetry_test:
        temperatures = star.symmetry_faces(star.temperatures)
        log_g = star.symmetry_faces(star.log_g)
    else:
        temperatures = star.temperatures
        log_g = star.log_g

    retval = {
        'star': atm.NaiveInterpolatedAtm.radiance(
                    **dict(
                        temperature=temperatures,
                        log_g=log_g,
                        metallicity=star.metallicity,
                        atlas=star.atmosphere or settings.ATM_ATLAS,
                        **kwargs
                    )
                )
    }

    if symmetry_test:
        retval['star'] = {band: star.mirror_face_values(vals) for band, vals in retval['star'].items()}

    return retval


def get_limbdarkening_cfs(system, **kwargs):
    """
    Returns limb darkening coefficients for each face.

    :param system: elisa.single_system.container.SystemContainer;
    :param kwargs: Dict;
    :**kwargs options**:
        * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
        * ** left_bandwidth ** * - float
        * ** right_bandwidth ** * - float
    :return: Dict[str, numpy.array];
    """
    star_container = system.star
    symmetry_test = not system.has_spots() and not system.has_pulsations()

    return {'star': get_component_limbdarkening_cfs(
        star_container,
        symmetry_test,
        kwargs['passband']
    )}




import numpy as np

from scipy import optimize
from copy import deepcopy

from . orbit import orbit
from . curves import lc, rv, c_router
from . transform import SingleSystemProperties
from . import (
    model,
    graphic,
    radius as sradius,
    utils as sys_utils
)
from . container import SinglePositionContainer
from .. base.types import INT
from .. logger import getLogger
from .. import const
from .. import (
    units,
    utils,
    const as c,
)
from .. base.system import System
from .. base.curves import utils as rv_utils
from .. opt.fsolver import fsolve
from .. base.star import Star

logger = getLogger('single_system.system')


class SingleSystem(System):
    """
    Class to store and calculate necessary properties of the single star system based on the user provided parameters.
    Child class of elisa.base.system.System.

    Class can be imported directly:
    ::

        from elisa import SingleSystem

    After initialization, apart from the attributes already defined by the user with the arguments, user has access to
    the following attributes:

        :angular_velocity: float; angular velocity of the stellar rotation

    `SingleSystem' requires instance of elisa.base.star.Star in `star` argument with the following
    mandatory arguments:

        :param mass: float; If mass is int, np.int, float, np.float, program assumes solar mass as it's unit.
                            If mass astropy.unit.quantity.Quantity instance, program converts it to default units.
        :param t_eff: float; Accepts value in any temperature unit. If your input is without unit,
                             function assumes that supplied value is in K.
        :param polar_log_g: float; log_10 of the polar surface gravity

    following optional arguments are also available:

        :param metallicity: float; log[M/H] default value is 0.0
        :param gravity_darkening: float; gravity darkening factor, if not supplied, it is interpolated from Claret 2003
                                         based on t_eff
        :param limb_darkening_coefficients: Union[float, dict]; optional limb darkening coefficients
                                            used for the whole star useful in case the modelled star is outside the
                                            supported range of atmospheric parameters. Limb darkening coefficients can
                                            be supplied as dict {passband: ld_coefs}. If unused, elisa will
                                            interpolate the values from supplied limb-darkening tables.

    Each component instance will after initialization contain following attributes:

        :critical_surface_potential: float; potential of the star required to fill its Roche lobe
        :equivalent_radius: float; radius of a sphere with the same volume as a component
        :polar_radius: float; radius of a star towards the pole of the star
        :equatorial_radius: float; radius of a star towards the pole of the star

    The SingleSystem can be initialized either by using valid class arguments, e.g.:
    ::

        from astropy import units as u

        from elisa.single_system.system import SingleSystem
        from elisa.base.star import Star

        star = Star(
            mass=1.0*u.solMass,
            t_eff=5772*u.K,
            gravity_darkening=0.32,
            polar_log_g=4.43775*u.dex(u.cm/u.s**2),
            metallicity=0.0,
            discretization_factor=2
        )
        
        system = SingleSystem(
            star=star,
            gamma=0*u.km/u.s,
            inclination=90*u.deg,
            rotation_period=25.380*u.d,
            reference_time=0.0*u.d,
            distance = 153*u.pc
        )

    or by using the SingleSystem.from_json(dict) function that accepts various parameter combination in form of
    dictionary such as:
    ::

        data = {
            "system": {
                "inclination": 90.0,
                "rotation_period": 10.1,
                "gamma": '10000 K',  # you can define quantity using a string representation of the astropy quantities
                "reference_time": 0.5,
                "phase_shift": 0.0,
                "distance": 64  # pc
            },
            "star": {
                "mass": 1.0,
                "t_eff": 5772.0,
                "gravity_darkening": 0.32,
                "discretization_factor": 5,
                "metallicity": 0.0,
                "polar_log_g": "4.43775 dex(cm.s-2)"   # you can use logarithmic units using json/dict input
            }
        }

        single = SingleSystem.from_json(data)

    See documentation for `from_json` method for details.

    The rotation of the binary system can be modelled using function
    `calculate_lines_of_sight(phases)`. E.g.:
    ::

        single_instance.calculate_lines_of_sight(np.linspace(0, 1))

    The class contains substantial plotting capability in the SingleSystem.plot module that contains following
    functions (further info in: see elisa.single_system.graphics.plot):

        - equipotential(args): zx cross-sections of equipotential surface
        - mesh(args): 3D mesh (scatter) plot of the surface points
        - wireframe(args): wire frame model of the star
        - surface(args): plot model of the star with various surface colormaps (gravity_acceleration,
          temperature, radiance, ...)

    Plot function can be called as function of the plot module. E.g.:
    ::

        single_instance.plot.surface(phase=0.1, colormap='temperature)

    Similarly, an animation of the orbital motion can be produced using SingleSystem.animation module and its function
    `rotational_motion(*args)`.

    List of valid system arguments:

    :param star: elisa.base.star.Star; instance of the single star
    :param inclination: Union[float, astropy.unit.quantity.Quantity]; Inclination of the system.
                        If unit is not supplied, value in degrees is assumed.
    :param rotational_period: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity]; Orbital period of
                              binary star system. If unit is not specified, default period unit is assumed (days).
    :param reference_time: Union[(numpy.)float, (numpy.)int, astropy.units.quantity.Quantity];
    :param phase_shift: float; Phase shift of the primary eclipse with respect to the ephemeris.
                               true_phase is used during calculations, where: true_phase = phase + phase_shift.
    :param additional_light: float; fraction of light that does not originate from the `BinarySystem`
    :param gamma: Union[float, astropy.unit.quantity.Quantity]; Center of mass velocity.
                  Expected type is astropy.units.quantity.Quantity, numpy.float or numpy.int
                  otherwise TypeError will be raised. If unit is not specified, default velocity unit is assumed (m/s).
    :param distance: float: distance between system and the observer
    """

    MANDATORY_KWARGS = ['inclination', 'rotation_period']
    OPTIONAL_KWARGS = ['reference_time', 'phase_shift', 'additional_light', 'gamma', 'distance']
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS

    STAR_MANDATORY_KWARGS = ['mass', 't_eff', 'polar_log_g']
    STAR_OPTIONAL_KWARGS = ['metallicity', 'gravity_darkening']
    STAR_ALL_KWARGS = STAR_MANDATORY_KWARGS + STAR_OPTIONAL_KWARGS

    def __init__(self, star, name=None, **kwargs):
        utils.invalid_kwarg_checker(kwargs, SingleSystem.ALL_KWARGS, self.__class__)
        utils.check_missing_kwargs(SingleSystem.MANDATORY_KWARGS, kwargs, instance_of=SingleSystem)
        self.object_params_validity_check(dict(star=star), self.STAR_MANDATORY_KWARGS)
        kwargs = self.transform_input(**kwargs)

        super(SingleSystem, self).__init__(name, **kwargs)

        logger.info(f"initialising object {self.__class__.__name__}")
        logger.debug(f"setting properties of a star in class instance {self.__class__.__name__}")

        self.plot = graphic.plot.Plot(self)
        self.animation = graphic.animation.Animation(self)

        self.star = star
        self._components = dict(star=self.star)

        # default values of properties
        self.orbit = None
        self.rotation_period = None
        self.reference_time = 0
        self.angular_velocity = None
        self.period = self.rotation_period
        self.phase_shift = 0.0

        # set attributes and test whether all parameters were initialized
        # we already ensured that all kwargs are valid and all mandatory kwargs are present so lets set class attributes
        self.init_properties(**kwargs)

        # calculation of dependent parameters
        self.angular_velocity = orbit.angular_velocity(self.rotation_period)
        self.star.surface_potential = model.surface_potential_from_polar_log_g(self.star.polar_log_g, self.star.mass)

        # orbit initialisation (initialise class Orbit from given BinarySystem parameters)
        self.init_orbit()

        self.setup_critical_potential()
        # checking if star is below break-up rotational velocity
        self.check_stability()

        # this is also check if star surface is closed
        self.setup_radii(calculate_equivalent_radius=True)
        self.setup_betas()
        self.assign_pulsations_amplitudes()
        self.setup_discretisation_factor()

        # setting common reference to emphemeris
        self.period = self.rotation_period
        self.t0 = self.reference_time

    @property
    def default_input_units(self):
        """
        Returns set of default units of intialization parameters, in case, when provided without units.

        :return: elisa.units.DefaultSingleSystemInputUnits;
        """
        return units.DefaultSingleSystemInputUnits

    @property
    def default_internal_units(self):
        """
        Returns set of internal units of system parameters.

        :return: elisa.units.DefaultSingleSystemUnits;
        """
        return units.DefaultSingleSystemUnits

    @classmethod
    def from_json(cls, data, _verify=True, _kind_of=None):
        """
        Create instance of BinarySystem from JSON in `standard` or `radius` format. Example of a `standard` format of
        parameters::

            {
                "system": {
                    "inclination": 90.0,
                    "rotation_period": 10.1,
                    "gamma": 10000,
                    "reference_time": 0.5,
                    "phase_shift": 0.0,
                    "distance": "452 pc"
                },
                "star": {
                    "mass": 1.0,
                    "t_eff": 5772.0,
                    "gravity_darkening": 0.32,
                    "discretization_factor": 5,
                    "metallicity": 0.0,
                    "polar_log_g": "4.43775 dex(cm.s-2)"   # you can also use logarithmic units using json/dict input
                }
            }

        Example of a `radius` format of parameters::

            {
                "system": {
                    "inclination": 90.0,
                    "rotation_period": 10.1,
                    "gamma": 10000,
                    "reference_time": 0.5,
                    "phase_shift": 0.0,
                    "distance": "984 pc"
                },
                "star": {
                    "mass": 1.0,
                    "t_eff": 5772.0,
                    "gravity_darkening": 0.32,
                    "discretization_factor": 5,
                    "metallicity": 0.0,
                    "equivalent_radius": "1.0 solRad"
                }
            }

        Default units::

             {
                "inclination": [degrees],
                "rotational_period": [days],
                "gamma": [m/s],
                "reference_time": [d],
                "phase_shift": [dimensionless],
                "distance": [pc],
                "mass": [solMass],
                "surface_potential": [dimensionless],
                "synchronicity": [dimensionless],
                "t_eff": [K],
                "gravity_darkening": [dimensionless],
                "discretization_factor": [degrees],
                "metallicity": [dimensionless],
                "semi_major_axis": [solRad],
                "mass_ratio": [dimensionless],
                "polar_log_g": [dex(m*s-2)],
                "equivalent_radius": [solRad],
                "limb_darkening_coefficients": [dimensionless]
            }

        :return: elisa.single_system.system.SingleSystem
        """
        data_cp = deepcopy(data)
        if _verify:
            sys_utils.validate_single_json(data_cp)

        kind_of = _kind_of or sys_utils.resolve_json_kind(data_cp)
        if kind_of in ["radius"]:
            data_cp = sys_utils.transform_json_radius_to_std(data_cp)

        star = Star(**data_cp["star"])
        return cls(star=star, **data_cp["system"])

    def build_container(self, phase=None, time=None, build_pulsations=True):
        """
        Function returns `OrbitalPositionContainer` with fully built model binary system at user-defined photometric
        phase or time of observation.

        :param time: float; JD
        :param phase: float; photometric phase
        :param build_pulsations: bool;
        :return: elisa.binary_system.container.OrbitalPositionContainer;
        """
        if phase is not None and time is not None:
            raise ValueError('Please specify whether you want to build your container EITHER at given photometric '
                             '`phase` or at given `time`.')
        phase = phase if time is None else utils.jd_to_phase(time, period=self.period, t0=self.reference_time)

        position = self.calculate_lines_of_sight(input_argument=phase, return_nparray=False, calculate_from='phase')[0]
        position_container = SinglePositionContainer.from_single_system(self, position)
        position_container.build(build_pulsations=build_pulsations)

        logger.info(f'Orbital position container was successfully built at photometric phase {phase:.2f}.')
        return position_container

    @classmethod
    def is_property(cls, kwargs):
        """
        Method for checking if keyword arguments are valid properties of this class.

        :param kwargs: Dict;
        """
        is_not = ['`{}`'.format(k) for k in kwargs if k not in cls.ALL_KWARGS]
        if is_not:
            raise AttributeError('Arguments {} are not valid {} properties.'.format(', '.join(is_not), cls.__name__))

    def critical_break_up_radius(self):
        """
        returns critical, break-up equatorial radius for given mass and rotational period

        :return: float
        """
        return np.power(c.G * self.star.mass / np.power(self.angular_velocity, 2), 1.0 / 3.0)

    def critical_break_up_velocity(self):
        """
        returns critical, break-up equatorial rotational velocity for given mass and rotational period

        :return: float
        """
        return np.power(c.G * self.star.mass * self.angular_velocity, 1.0 / 3.0)

    def get_info(self):
        pass

    # _________________________AFTER_REFACTOR___________________________________

    def init(self):
        """
        function to reinitialize SingleSystem class instance after changing parameter(s) of binary system using setters

        :return:
        """
        logger.info(f're/initialising class instance {SingleSystem.__name__}')
        self.__init__(star=self.star, **self.kwargs_serializer())

    def init_orbit(self):
        """
        Initialization of single system orbit class. Similar in functionality of Orbit in Binary system. Simulates
        apparent motion of the observer around single star as the star rotates.
        """
        logger.debug(f"re/initializing orbit in class instance {self.__class__.__name__} / {self.name}")
        orbit_kwargs = {key: getattr(self, key) for key in orbit.Orbit.ALL_KWARGS}
        self.orbit = orbit.Orbit(**orbit_kwargs)

    def setup_discretisation_factor(self):
        if self.star.has_spots():
            for spot in self.star.spots.values():
                if not spot.kwargs.get('discretization_factor'):
                    spot.discretization_factor = self.star.discretization_factor

    @staticmethod
    def is_eccentric():
        """
        Resolve whether system is eccentric.

        :return: bool;
        """
        return False

    def calculate_radii(self):
        """
        Calculates important radii.

        :return: Dict;
        """
        fns = [sradius.calculate_polar_radius, sradius.calculate_equatorial_radius]
        radii = dict(star=dict())
        for fn in fns:
            logger.debug(f'initialising {" ".join(str(fn.__name__).split("_")[1:])} for the star')
            param = f'{"_".join(str(fn.__name__).split("_")[1:])}'
            kwargs = dict(mass=self.star.mass,
                          angular_velocity=self.angular_velocity,
                          surface_potential=self.star.surface_potential)
            try:
                r = fn(**kwargs)
            except Exception as e:
                raise ValueError(f'Function {fn.__name__} was not able to calculate its radius. '
                                 f'Your system is not physical. Exception: {str(e)}')
            radii['star'][param] = r

        return radii

    def setup_radii(self, calculate_equivalent_radius=True):
        """
        Auxiliary function for calculation of important radii.
        """
        radii = self.calculate_radii()
        instance: Star = getattr(self, 'star')

        for key, value in radii['star'].items():
            setattr(instance, key, value)

        if calculate_equivalent_radius:
            setattr(instance, 'equivalent_radius', self.calculate_equivalent_radius())

    @property
    def components(self):
        """
        Return components object in Dict.

        :return: Dict[str, elisa.base.Star]
        """
        return self._components

    def calculate_equipotential_boundary(self):
        """
        calculates a equipotential boundary of star in zx(yz) plane

        :return: Tuple; (np.array, np.array)
        """
        points = []
        angles = np.linspace(0, c.FULL_ARC, 300, endpoint=True)
        init_val = - c.G * self.star.mass / self.star.surface_potential
        scipy_solver_init_value = np.array([init_val])

        for angle in angles:
            precalc_args = (self.star.mass, self.angular_velocity, angle)
            argss = (model.pre_calculate_for_potential_value(*precalc_args), self.star.surface_potential)
            solution, _, ier, _ = fsolve(model.potential_fn, scipy_solver_init_value, full_output=True, args=argss)
            if ier == 1 and not np.isnan(solution[0]):
                solution = solution[0]
            else:
                continue

            points.append([solution * np.sin(angle), solution * np.cos(angle)])
        return np.array(points)

    def properties_serializer(self):
        props = SingleSystemProperties.transform_input(**self.kwargs_serializer())
        props.update({
            "angular_velocity": self.angular_velocity,
        })
        return props

    def transform_input(self, **kwargs):
        """
        Transform and validate input kwargs.
        :param kwargs: Dict
        :return: Dict
        """
        return SingleSystemProperties.transform_input(**kwargs)

    def setup_critical_potential(self):
        self.star.critical_surface_potential = self.calculate_critical_potential()

    def calculate_critical_potential(self):
        """
        Compute and set critical surface potential.
        Critical surface potential is potential when component is stable for give mass and rotaion period.
        """
        precalc_args = self.star.mass, self.angular_velocity, c.HALF_PI
        args = (model.pre_calculate_for_potential_value(*precalc_args), 0.0)

        x0 = - c.G * self.star.mass / self.star.surface_potential
        solution = optimize.newton(model.radial_potential_derivative, x0, args=args[0], tol=1e-12)
        if np.isnan(solution):
            raise ValueError("Iteration process to solve critical potential seems "
                             "to lead nowhere (critical potential solver has failed).")
        return model.potential_fn(solution, *args)

    def check_stability(self):
        """
        Checks if star is rotationally stable.
        """
        if self.star.critical_surface_potential < self.star.surface_potential:
            raise ValueError('Non-physical system. Star rotation is above critical break-up velocity.')

    def get_positions_method(self):
        return self.calculate_lines_of_sight

    def calculate_lines_of_sight(self, input_argument=None, return_nparray=False, calculate_from='phase'):
        """
        Returns vectors oriented in direction star -> observer for given set of phases or azimuths.

        :param calculate_from: str; 'phase' or 'azimuths' parameter based on which the orbital motion should be
                                    calculated
        :param return_nparray: bool; if True positions in form of numpy arrays will be also returned
        :param input_argument: numpy.array;
        :return: Tuple[List[NamedTuple: elisa.const.Position], List[Integer]] or
                 List[NamedTuple: elisa.const.Position]
        """
        input_argument = np.array([input_argument]) if np.isscalar(input_argument) else input_argument
        rotational_motion = self.orbit.rotational_motion(phase=input_argument) if calculate_from == 'phase' \
            else self.orbit.rotational_motion_from_azimuths(azimuth=input_argument)
        idx = np.arange(np.shape(input_argument)[0], dtype=INT)[:, np.newaxis]
        positions = np.hstack((idx, np.full(idx.shape, np.nan), rotational_motion))

        return positions if return_nparray else [const.Position(*p) for p in positions]

    def calculate_equivalent_radius(self):
        """
        Function returns equivalent radius of the star, i.e. radius of the sphere with the same volume as
        given component.

        :return: float;
        """
        volume = sys_utils.calculate_volume(self)
        return utils.calculate_equiv_radius(volume)

    # light curves *****************************************************************************************************
    def compute_lightcurve(self, **kwargs):
        """
        This function decides which light curve generator function is used.
        Depending on the basic properties of the binary system.

        :param kwargs: Dict; arguments to be passed into light curve generator functions
            * ** passband ** * - Dict[str, elisa.observer.PassbandContainer]
            * ** left_bandwidth ** * - float
            * ** right_bandwidth ** * - float
            * ** phases ** * - numpy.array
            * ** position_method ** * - method
        :return: Dict; {`passband`: numpy.array, }
        """
        fn_arr = (self._compute_light_curve_without_pulsations, self._compute_light_curve_with_pulsations)

        curve_fn = c_router.resolve_curve_method(self, fn_arr)
        return curve_fn(**kwargs)

    def _compute_light_curve_with_pulsations(self, **kwargs):
        return lc.compute_light_curve_with_pulsations(self, **kwargs)

    def _compute_light_curve_without_pulsations(self, **kwargs):
        return lc.compute_light_curve_without_pulsations(self, **kwargs)

    # radial velocity curves *******************************************************************************************
    def compute_rv(self, **kwargs):
        """
        This function decides which radial velocities generator function is
        used depending on the user-defined method.

        :param kwargs: Dict;
        :**kwargs options**:
            * :method: str; `kinematic` (motion of the centre of mass) or
                            `radiometric` (radiance weighted contribution of each visible element)
            * :position_method: callable; method for obtaining orientation of the star
            * :phases: numpy.array; photometric phases

        :return: Dict; {`star`: numpy.array}
        """
        if kwargs['method'] == 'kinematic':
            return rv.com_radial_velocity(self, **kwargs)
        if kwargs['method'] == 'radiometric':
            fn_arr = (self._compute_rv_curve_without_pulsations,
                      self._compute_rv_curve_with_pulsations)
            curve_fn = c_router.resolve_curve_method(self, fn_arr)

            kwargs = rv_utils.include_passband_data_to_kwargs(**kwargs)
            return curve_fn(**kwargs)

    def _compute_rv_curve_with_pulsations(self, **kwargs):
        return rv.compute_rv_curve_with_pulsations(self, **kwargs)

    def _compute_rv_curve_without_pulsations(self, **kwargs):
        return rv.compute_rv_curve_without_pulsations(self, **kwargs)



import numpy as np

from . import model
from .. import (
    const,
    umpy as up
)
from .. opt.fsolver import fsolve


def calculate_radius(synchronicity, mass_ratio, surface_potential, component, *args):
    """
    Function calculates radius of the star in given direction of arbitrary direction vector (in spherical
    coordinates) starting from the centre of the star.

    :param component: str; `primary` or `secondary`,
    :param synchronicity: float;
    :param mass_ratio: float;
    :param surface_potential: float; if None compoent surface potential is assumed
    :param args: Tuple;

    ::

        (
            components_distance: float - distance between components in SMA units,
            phi: float - longitudonal angle of direction vector measured from point under L_1 in
                         positive direction (in radians)
            theta: float - latitudonal angle of direction vector measured from north pole (in radians)
         )

    :return: float; radius
    """
    if component == 'primary':
        fn = model.potential_primary_fn
        precalc = model.pre_calculate_for_potential_value_primary
    elif component == 'secondary':
        fn = model.potential_secondary_fn
        precalc = model.pre_calculate_for_potential_value_secondary
    else:
        raise ValueError(f'Invalid value of `component` argument {component}. \n'
                         f'Expecting `primary` or `secondary`.')

    precalc_args = (synchronicity, mass_ratio) + args
    scipy_solver_init_value = np.array([1e-4])
    argss = ((mass_ratio,) + precalc(*precalc_args), surface_potential)
    solution, _, ier, _ = fsolve(fn, scipy_solver_init_value, full_output=True, args=argss, xtol=1e-10)

    # check for regular solution
    if ier == 1 and not up.isnan(solution[0]) and 30 >= solution[0] >= 0:
        return solution[0]
    else:
        if not (0 < solution[0] < 1.0):
            raise ValueError(f'Invalid value of radius {solution} was calculated.')
        return solution[0]


def calculate_polar_radius(synchronicity, mass_ratio, components_distance, surface_potential, component):
    """
    Radius of the star in the direction of the pole.

    :param synchronicity: float;
    :param mass_ratio: float;
    :param components_distance: float;
    :param surface_potential: float;
    :param component: str; `primary` a `secondary`
    :return: float;
    """
    args = (components_distance, 0.0, 0.0)
    return calculate_radius(synchronicity, mass_ratio, surface_potential, component, *args)


def calculate_side_radius(synchronicity, mass_ratio, components_distance, surface_potential, component):
    """
    Radius of the star in the direction perpendicular to the pole and component join vector.

    :param synchronicity: float;
    :param mass_ratio: float;
    :param components_distance: float;
    :param surface_potential: float;
    :param component: str; `primary` a `secondary`
    :return: float;
    """
    args = (components_distance, const.HALF_PI, const.HALF_PI)
    return calculate_radius(synchronicity, mass_ratio, surface_potential, component, *args)


def calculate_backward_radius(synchronicity, mass_ratio, components_distance, surface_potential, component):
    """
    Radius of the star in the direction away from the companion.

    :param synchronicity: float;
    :param mass_ratio: float;
    :param components_distance: float;
    :param surface_potential: float;
    :param component: str; `primary` a `secondary`
    :return: float;
    """
    args = (components_distance, const.PI, const.HALF_PI)
    return calculate_radius(synchronicity, mass_ratio, surface_potential, component, *args)


def calculate_forward_radius(synchronicity, mass_ratio, components_distance, surface_potential, component):
    """
    Radius of the star in the direction towards the companion.

    :param synchronicity: float;
    :param mass_ratio: float;
    :param components_distance: float;
    :param surface_potential: float;
    :param component: str; `primary` a `secondary`
    :return: float;
    """
    args = (components_distance, 0.0, const.HALF_PI)
    return calculate_radius(synchronicity, mass_ratio, surface_potential, component, *args)


def calculate_forward_radii(distances, surface_potential, mass_ratio, synchronicity, component):
    """
    Calculates forward radii for given object for given array of distances.

    :param distances: Union[numpy.array, List]: array of component distances at which to calculate
                      the forward radii of given component(s)
    :param surface_potential: float;
    :param mass_ratio: float;
    :param synchronicity: float;
    :param component: str;
    :return: Dict; Dict[str, numpy.array];
    """
    return [calculate_forward_radius(synchronicity, mass_ratio, d, surface_potential[ii], component)
            for ii, d in enumerate(distances)]



import gc
import numpy as np

from copy import copy

from . types import INT
from .. base.transform import SpotProperties
from .. import units as u, settings
from .. utils import is_empty
from .. logger import getLogger
from .. import (
    utils,
    umpy as up
)

logger = getLogger("base.spots")


class Spot(object):
    """
    Spot object container. It is available as a list item of a `Star.spots` attribute after the initialization of the
    host system. This spot class is producing a circular spot at a specified coordinates with a temperature difference
    between the spot and the host star described by the `temperature_factor` defined as
    t_eff,spot/t_eff,star.


    Input parameters:

    :param longitude: Union[(numpy.)int, (numpy.)float, astropy.unit.quantity.Quantity];
                      Expecting value in degrees or as astropy units instance.
    :param latitude: Union[(numpy.)int, (numpy.)float, astropy.unit.quantity.Quantity];
                     Expecting value in degrees or as astropy units instance.
    :param angular_radius: Union[(numpy.)int, (numpy.)float, astropy.unit.quantity.Quantity];
            `              Expecting value in degrees or as astropy units instance.
    :param temperature_factor: Union[(numpy.)int, (numpy.)float];
    :param discretization_factor: Union[(numpy.)int, (numpy.)float, astropy.unit.quantity.Quantity];
                                  Spot discretization_factor (mean angular size of spot face).
                                  Expecting value in degrees or as astropy units instance.

    Output parameters that describing the spot:
    
    :boundary: numpy.array; points at the spot/star
    :boundary_center: float; geometrical centre of the spot boundary
    :center: float; coordinates of the spot centre
    :points: numpy.array; surface points belonging to spot
    :normals: numpy.array; outward facing normal vectors of surface elements belonging to the spot
    :faces: numpy.array; surface element (triangle) simplices
    :face_centres: numpy.array; centers of spot surface elements
    :areas: numpy.array; areas of spot surface elements
    :potential_gradient_magnitudes: numpy.array;
    :temperatures: numpy.array; effective temperatures of surface elements
    :log_g: numpy.array; surface gravity distribution over the surface elements

    """
    MANDATORY_KWARGS = ["longitude", "latitude", "angular_radius", "temperature_factor"]
    OPTIONAL_KWARGS = ["discretization_factor"]
    ALL_KWARGS = MANDATORY_KWARGS + OPTIONAL_KWARGS

    def __init__(self, **kwargs):
        utils.invalid_kwarg_checker(kwargs=kwargs, kwarglist=Spot.ALL_KWARGS, instance=Spot)
        utils.check_missing_kwargs(Spot.MANDATORY_KWARGS, kwargs, instance_of=Spot)
        self.kwargs = self.transform_input(**kwargs)

        # supplied parameters
        self.discretization_factor = np.nan
        self.latitude = np.nan
        self.longitude = np.nan
        self.angular_radius = np.nan
        self.temperature_factor = np.nan

        # container parameters
        self.boundary = np.array([])
        self.boundary_center = np.nan
        self.center = np.nan

        self.points = np.array([])
        self.normals = np.array([])
        self.faces = np.array([])
        self.face_centres = np.array([])
        self.points_spherical = np.array([])

        self.velocities = np.array([])

        self.areas = np.array([])
        self.potential_gradient_magnitudes = np.array([])
        self.temperatures = np.array([])
        self.log_g = np.array([])

        self.init_properties(**self.kwargs)

    @property
    def default_input_units(self):
        """
        Returns set of default units of intialization parameters, in case, when provided without units.

        :return: elisa.units.DefaultSpotInputUnits;
        """
        return u.DefaultSpotInputUnits

    @property
    def default_internal_units(self):
        """
        Returns set of internal units of Spot parameters.

        :return: elisa.units.DefaultSpotUnits;
        """
        return u.DefaultSpotUnits

    @staticmethod
    def transform_input(**kwargs):
        return SpotProperties.transform_input(**kwargs)

    def calculate_areas(self):
        """
        Returns areas of each face of the spot build_surface.

        :return: numpy.array:

        ::

            numpy.array([area_1, ..., area_n])
        """
        return utils.triangle_areas(triangles=self.faces, points=self.points)

    def init_properties(self, **kwargs):
        for key in kwargs:
            set_val = kwargs.get(key)
            setattr(self, key, set_val)

    def kwargs_serializer(self):
        """
        Serializer and return mandatory kwargs of sefl (Spot) instance to dict.

        :return: Dict; { kwarg: value }
        """

        default_units = {
            "longitude": u.DefaultSpotUnits.longitude,
            "latitude": u.DefaultSpotUnits.latitude,
            "angular_radius": u.DefaultSpotUnits.angular_radius,
            "discretization_factor": u.DefaultSpotUnits.discretization_factor
        }

        serialized_kwargs = dict()
        for kwarg in self.ALL_KWARGS:
            if kwarg in default_units:
                value = getattr(self, kwarg)
                if not isinstance(value, u.Quantity):
                    value = value * default_units[kwarg]
                serialized_kwargs[kwarg] = value
            else:
                serialized_kwargs[kwarg] = getattr(self, kwarg)
        return serialized_kwargs


def split_points_of_spots_and_component(on_container, points, vertices_map):
    """
    Based on vertices map, separates points to points which belong to Star object
    and points which belong to each defined Spot object.
    During the process remove overlapped spots.

    :param on_container: Union; instance of object to split spots on
    :param points: numpy.array; all points of object (spot points and component points together)
    :param vertices_map: Union[List, numpy.array]; map which define refrences of index in
                         given Iterable to object (Spot or Star).
    :return: Dict;

    ::

        {
            "object": numpy.array([[pointN_x, pointN_y, pointN_z], ...]),
            "spot_index": numpy.array([[pointM_x, pointM_y, pointM_z], ...]), ...
        }
    """
    points = np.array(points)
    # component_points = {"object": points[up.where(np.array(vertices_map) == {'enum': -1})[0]]}
    component_points = {"object": points[vertices_map == -1]}
    # indices = set([int(val["enum"]) for val in vertices_map if val["enum"] > -1])
    indices = np.unique(vertices_map[vertices_map > -1])
    remove_overlaped_spots_by_spot_index(on_container, indices)
    # spots_points = {
    #     f"{i}": points[up.where(np.array(vertices_map) == {'enum': i})[0]] for i in range(len(on_container.spots))
    #     if len(up.where(np.array(vertices_map) == {'enum': i})[0]) > 0
    # }
    spots_points = {
        f"{i}": points[vertices_map == i] for i in range(len(on_container.spots))
        if len(vertices_map[vertices_map == i]) > 0
    }
    return {**component_points, **spots_points}


def setup_body_points(on_container, points):
    """
    Setup points for Star instance and spots based on input `points` Dict object.
    Such `points` map looks like following

    :param on_container: Union; instance to setup spot points and body points on
    :param points: Dict[str, numpy.array];

    ::

        {
            "object": [[point0_x, point0_y, point0_z], ..., [pointN_x, pointN_y, pointN_z]],
            "0": [<points>],
            "1": [<points>]...
        },

    where `object` contain numpy.array of object points and indices points for given spot.
    """
    on_container.points = points.pop("object")
    for spot_index, spot_points in points.items():
        on_container.spots[int(spot_index)].points = points[spot_index]


def incorporate_spots_mesh(to_container, component_com):
    """
    Based on spots definitions, evaluate spot points on Star surface and remove those points of Star itself
    which are inside of any spots. Do the same operation with iteratively for each Spot instance defined previously.
    Evaluation is running from index 0, what means that spot with lower index
    is overwriten by spot with higher index if located on top of each other.

    All points are assigned to given object (Star points to Star object and Spot points to related Spot object).

    Function defines variable `vertices_map` used in other methods. Structure of this variable is following::

        [-1, -1, -1, ..., 0, ..., 1, ...]

    Value -1 means that points on given
    index position of vertices_map belongs to Star point.
    Enum indices >= 0 means that given point belongs to Spot with given index.

    :param to_container: Union; instace to incorporate spots into
    :param component_com: float; center of mass of component (it's x coordinate)
    :return: to_container: Union; instace to incorporate spots into
    """
    if not to_container.spots:
        logger.debug(f"not spots found, skipping incorporating spots "
                     f"to_container mesh on component {to_container.name}")
        return to_container
    logger.debug(f"incorporating spot points to_container component {to_container.name} mesh")

    # vertices_map = [{"enum": -1} for _ in to_container.points]
    vertices_map = np.full(to_container.points.shape[0], -1)
    # `all_component_points` do not contain points of Any spot
    all_component_points = copy(to_container.points)

    neck = np.max(np.abs(to_container.points[:, 0] - component_com))

    for spot_index, spot in to_container.spots.items():
        # average spacing in spot points
        cos_max_angle_point = up.cos(spot.angular_radius + 0.30 * spot.discretization_factor)
        spot_center = spot.center - np.array([component_com, 0., 0.])

        # removing star points in spot
        # all component points means just points of component NOT merged points + spots

        com_pts = all_component_points - np.array([component_com, 0., 0.])[None, :]
        cos_angles = np.sum(np.multiply(spot_center[None, :], com_pts), axis=1) / (
                    np.linalg.norm(spot_center) * np.linalg.norm(com_pts, axis=1))

        # skip all points of object outside of spot
        angular_dist_cond = cos_angles < cos_max_angle_point
        # skip points near neck
        neck_cond = np.abs(np.abs(com_pts[:, 0]) - neck) < 1e-9
        in_condition = np.logical_or(angular_dist_cond, neck_cond)

        vertices_to_keep = np.arange(com_pts.shape[0], dtype=INT)[in_condition]

        # simplices of target object for testing whether point lying inside or not of spot boundary, removing
        # duplicate points on the spot border
        # vertices_to_remove = np.unique(vertices_to_remove)
        vertices_to_keep = np.unique(vertices_to_keep)
        all_component_points = all_component_points[vertices_to_keep]
        vertices_map = vertices_map[vertices_to_keep]

        all_component_points = np.row_stack((all_component_points, spot.points))
        vertices_map = np.concatenate((vertices_map, np.full(spot.points.shape[0], spot_index)))

    separated_points = split_points_of_spots_and_component(to_container, all_component_points, vertices_map)
    setup_body_points(to_container, separated_points)
    return to_container


def remap_surface_elements(on_container, mapper, points_to_remap):
    """
    Function remaps all surface points (`points_to_remap`) and faces (star and spots) according to the `model`.

    :param on_container: Union; container object with spots
    :param mapper: List; list of indices of points in `points_to_remap` divided into star and spots sublists
    :param points_to_remap: numpy.array; array of all surface points (star + points used in
                            `_split_spots_and_component_faces`)
    :return: on_container: Union; container object with spots
    """

    # remapping points and faces of star
    logger.debug(f"changing value of parameter points of component {on_container.name}")
    indices = np.unique(mapper["object"])
    on_container.points = points_to_remap[indices]

    logger.debug(f"changing value of parameter faces of component {on_container.name}")

    points_length = np.shape(points_to_remap)[0]
    remap_list = np.empty(points_length, dtype=INT)
    remap_list[indices] = up.arange(np.shape(indices)[0])
    on_container.faces = remap_list[mapper["object"]]

    # remapping points and faces of spots
    for spot_index, _ in list(on_container.spots.items()):
        logger.debug(f"changing value of parameter points of spot {spot_index} / component {on_container.name}")
        # get points currently belong to the given spot
        indices = np.unique(mapper["spots"][spot_index])
        on_container.spots[spot_index].points = points_to_remap[indices]

        logger.debug(f"changing value of parameter faces of spot {spot_index} / component {on_container.name}")

        remap_list = np.empty(points_length, dtype=INT)
        remap_list[indices] = up.arange(np.shape(indices)[0])
        on_container.spots[spot_index].faces = remap_list[mapper["spots"][spot_index]]
    gc.collect()
    return on_container


def remove_overlaped_spots_by_spot_index(from_container, keep_spot_indices, _raise=True):
    """
    Remove definition and instance of those spots that are overlaped
    by another one and basically has no face to work with.

    :param from_container: Union; container object with spots
    :param keep_spot_indices: List[int]; list of spot indices to keep
    :param _raise: bool;
    :return: from_container: Union; container object with spots
    """
    all_spot_indices = set([int(val) for val in from_container.spots.keys()])
    spot_indices_to_remove = all_spot_indices.difference(keep_spot_indices)
    spots_meta = [from_container.spots[idx].kwargs_serializer()
                  for idx in from_container.spots if idx in spot_indices_to_remove]
    spots_meta = '\n'.join([str(meta) for meta in spots_meta])
    if _raise and not is_empty(spot_indices_to_remove):
        raise ValueError(f"Spots {spots_meta} have no pointns to continue.\nPlease, specify spots wisely.")
    for spot_index in spot_indices_to_remove:
        from_container.remove_spot(spot_index)
    return from_container


def remove_overlaped_spots_by_vertex_map(from_container, vertices_map):
    """
    Remove spots of Start object that are totally overlapped by another spot.

    :param from_container: Union; container object with spots
    :param vertices_map: Union[List, numpy.array]
    :return: from_container: Union; container object with spots
    """
    # remove spots that are totaly overlaped
    spots_instance_indices = list(set([vertices_map[ix]["enum"] for ix, _ in enumerate(vertices_map)
                                       if vertices_map[ix]["enum"] >= 0]))
    for spot_index, _ in list(from_container.spots.items()):
        if spot_index not in spots_instance_indices:
            if not settings.SUPPRESS_WARNINGS:
                logger.warning(f"spot with index {spot_index} doesn't contain Any face "
                               f"and will be removed from component {from_container.name} spot list")
            from_container.remove_spot(spot_index=spot_index)
    gc.collect()
    return from_container
